# -*- coding: utf-8 -*-
# flake8: noqa: F401
from importlib.metadata import version


__version__ = version(__name__)


import nest_asyncio

nest_asyncio.apply()

from chatsky.pipeline import Pipeline
from chatsky.script import Context, Script

from chatsky.msg import *

Script.model_rebuild()

import chatsky.__rebuild_pydantic_models__



# -*- coding: utf-8 -*-
# flake8: noqa: F401

from chatsky.slots.slots import GroupSlot, ValueSlot, RegexpSlot, FunctionSlot
from chatsky.slots.conditions import slots_extracted
from chatsky.slots.processing import extract, extract_all, unset, unset_all, fill_template
from chatsky.slots.response import filled_template



"""
Response
---------------------------
Slot-related Chatsky responses.
"""

from __future__ import annotations
from typing import Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from chatsky.script import Context, Message
    from chatsky.pipeline import Pipeline


def filled_template(template: Message) -> Callable[[Context, Pipeline], Message]:
    """
    Fill template with slot values.
    The `text` attribute of the template message should be a format-string:
    e.g. "Your username is {profile.username}".

    For the example above, if ``profile.username`` slot has value "admin",
    it would return a copy of the message with the following text:
    "Your username is admin".

    :param template: Template message with a format-string text.
    """

    def fill_inner(ctx: Context, pipeline: Pipeline) -> Message:
        message = template.model_copy()
        new_text = ctx.framework_data.slot_manager.fill_template(template.text)
        message.text = new_text
        return message

    return fill_inner



# -*- coding: utf-8 -*-
from .common import is_interactive_mode, check_happy_path, run_interactive_mode
from .toy_script import TOY_SCRIPT, TOY_SCRIPT_ARGS, HAPPY_PATH
from .response_comparers import default_comparer

try:
    import pytest

    pytest.register_assert_rewrite("chatsky.utils.testing.telegram")
except ImportError:
    ...



"""
Devel Utils
-----------
These utils contain useful classes/functions that are often used in various
parts of the framework.
"""

from .json_serialization import (
    JSONSerializableDict,
    PickleEncodedValue,
    JSONSerializableExtras,
)
from .extra_field_helpers import grab_extra_fields
from .async_helpers import wrap_sync_function_in_async



# -*- coding: utf-8 -*-

from .database import DBContextStorage, threadsafe_method, context_storage_factory
from .json import JSONContextStorage, json_available
from .pickle import PickleContextStorage, pickle_available
from .sql import SQLContextStorage, postgres_available, mysql_available, sqlite_available, sqlalchemy_available
from .ydb import YDBContextStorage, ydb_available
from .redis import RedisContextStorage, redis_available
from .mongo import MongoContextStorage, mongo_available
from .shelve import ShelveContextStorage
from .protocol import PROTOCOLS, get_protocol_install_suggestion



"""
Component
---------
The Component module defines a :py:class:`.PipelineComponent` class,
which is a fundamental building block of the framework. A PipelineComponent represents a single
step in a processing pipeline, and is responsible for performing a specific task or set of tasks.

The PipelineComponent class can be a group or a service. It is designed to be reusable and composable,
allowing developers to create complex processing pipelines by combining multiple components.
"""

from __future__ import annotations
import logging
import abc
import asyncio
from typing import Optional, Awaitable, TYPE_CHECKING

from chatsky.script import Context

from ..service.extra import BeforeHandler, AfterHandler
from ..conditions import always_start_condition
from ..types import (
    StartConditionCheckerFunction,
    ComponentExecutionState,
    ServiceRuntimeInfo,
    GlobalExtraHandlerType,
    ExtraHandlerFunction,
    ExtraHandlerType,
    ExtraHandlerBuilder,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from chatsky.pipeline.pipeline.pipeline import Pipeline


class PipelineComponent(abc.ABC):
    """
    This class represents a pipeline component, which is a service or a service group.
    It contains some fields that they have in common.

    :param before_handler: :py:class:`~.BeforeHandler`, associated with this component.
    :type before_handler: Optional[:py:data:`~.ExtraHandlerBuilder`]
    :param after_handler: :py:class:`~.AfterHandler`, associated with this component.
    :type after_handler: Optional[:py:data:`~.ExtraHandlerBuilder`]
    :param timeout: (for asynchronous only!) Maximum component execution time (in seconds),
        if it exceeds this time, it is interrupted.
    :param requested_async_flag: Requested asynchronous property;
        if not defined, `calculated_async_flag` is used instead.
    :param calculated_async_flag: Whether the component can be asynchronous or not
        1) for :py:class:`~.pipeline.service.service.Service`: whether its `handler` is asynchronous or not,
        2) for :py:class:`~.pipeline.service.group.ServiceGroup`: whether all its `services` are asynchronous or not.

    :param start_condition: StartConditionCheckerFunction that is invoked before each component execution;
        component is executed only if it returns `True`.
    :type start_condition: Optional[:py:data:`~.StartConditionCheckerFunction`]
    :param name: Component name (should be unique in single :py:class:`~.pipeline.service.group.ServiceGroup`),
        should not be blank or contain `.` symbol.
    :param path: Separated by dots path to component, is universally unique.
    """

    def __init__(
        self,
        before_handler: Optional[ExtraHandlerBuilder] = None,
        after_handler: Optional[ExtraHandlerBuilder] = None,
        timeout: Optional[float] = None,
        requested_async_flag: Optional[bool] = None,
        calculated_async_flag: bool = False,
        start_condition: Optional[StartConditionCheckerFunction] = None,
        name: Optional[str] = None,
        path: Optional[str] = None,
    ):
        self.timeout = timeout
        """
        Maximum component execution time (in seconds),
        if it exceeds this time, it is interrupted (for asynchronous only!).
        """
        self.requested_async_flag = requested_async_flag
        """Requested asynchronous property; if not defined, :py:attr:`~requested_async_flag` is used instead."""
        self.calculated_async_flag = calculated_async_flag
        """Calculated asynchronous property, whether the component can be asynchronous or not."""
        self.start_condition = always_start_condition if start_condition is None else start_condition
        """
        Component start condition that is invoked before each component execution;
        component is executed only if it returns `True`.
        """
        self.name = name
        """
        Component name (should be unique in single :py:class:`~pipeline.service.group.ServiceGroup`),
        should not be blank or contain '.' symbol.
        """
        self.path = path
        """
        Dot-separated path to component (is universally unique).
        This attribute is set in :py:func:`~chatsky.pipeline.pipeline.utils.finalize_service_group`.
        """

        self.before_handler = BeforeHandler([] if before_handler is None else before_handler)
        self.after_handler = AfterHandler([] if after_handler is None else after_handler)

        if name is not None and (name == "" or "." in name):
            raise Exception(f"User defined service name shouldn't be blank or contain '.' (service: {name})!")

        if not calculated_async_flag and requested_async_flag:
            raise Exception(f"{type(self).__name__} '{name}' can't be asynchronous!")

    def _set_state(self, ctx: Context, value: ComponentExecutionState):
        """
        Method for component runtime state setting, state is preserved in `ctx.framework_data`.

        :param ctx: :py:class:`~.Context` to keep state in.
        :param value: State to set.
        :return: `None`
        """
        ctx.framework_data.service_states[self.path] = value

    def get_state(self, ctx: Context, default: Optional[ComponentExecutionState] = None) -> ComponentExecutionState:
        """
        Method for component runtime state getting, state is preserved in `ctx.framework_data`.

        :param ctx: :py:class:`~.Context` to get state from.
        :param default: Default to return if no record found
            (usually it's :py:attr:`~.pipeline.types.ComponentExecutionState.NOT_RUN`).
        :return: :py:class:`~pipeline.types.ComponentExecutionState` of this service or default if not found.
        """
        return ctx.framework_data.service_states.get(self.path, default if default is not None else None)

    @property
    def asynchronous(self) -> bool:
        """
        Property, that indicates, whether this component is synchronous or asynchronous.
        It is calculated according to the following rules:

        - | If component **can** be asynchronous and :py:attr:`~requested_async_flag` is set,
            it returns :py:attr:`~requested_async_flag`.
        - | If component **can** be asynchronous and :py:attr:`~requested_async_flag` isn't set,
            it returns `True`.
        - | If component **can't** be asynchronous and :py:attr:`~requested_async_flag` is `False` or not set,
            it returns `False`.
        - | If component **can't** be asynchronous and :py:attr:`~requested_async_flag` is `True`,
            an Exception is thrown in constructor.

        """
        return self.calculated_async_flag if self.requested_async_flag is None else self.requested_async_flag

    async def run_extra_handler(self, stage: ExtraHandlerType, ctx: Context, pipeline: Pipeline):
        extra_handler = None
        if stage == ExtraHandlerType.BEFORE:
            extra_handler = self.before_handler
        if stage == ExtraHandlerType.AFTER:
            extra_handler = self.after_handler
        if extra_handler is None:
            return
        try:
            extra_handler_result = await extra_handler(ctx, pipeline, self._get_runtime_info(ctx))
            if extra_handler.asynchronous and isinstance(extra_handler_result, Awaitable):
                await extra_handler_result
        except asyncio.TimeoutError:
            logger.warning(f"{type(self).__name__} '{self.name}' {extra_handler.stage} extra handler timed out!")

    @abc.abstractmethod
    async def _run(self, ctx: Context, pipeline: Pipeline) -> None:
        """
        A method for running pipeline component, it is overridden in all its children.
        This method is run after the component's timeout is set (if needed).

        :param ctx: Current dialog :py:class:`~.Context`.
        :param pipeline: This :py:class:`~.Pipeline`.
        """
        raise NotImplementedError

    async def __call__(self, ctx: Context, pipeline: Pipeline) -> Optional[Awaitable]:
        """
        A method for calling pipeline components.
        It sets up timeout if this component is asynchronous and executes it using :py:meth:`~._run` method.

        :param ctx: Current dialog :py:class:`~.Context`.
        :param pipeline: This :py:class:`~.Pipeline`.
        :return: `None` if the service is synchronous; an `Awaitable` otherwise.
        """
        if self.asynchronous:
            task = asyncio.create_task(self._run(ctx, pipeline))
            return asyncio.wait_for(task, timeout=self.timeout)
        else:
            return await self._run(ctx, pipeline)

    def add_extra_handler(self, global_extra_handler_type: GlobalExtraHandlerType, extra_handler: ExtraHandlerFunction):
        """
        Method for adding a global extra handler to this particular component.

        :param global_extra_handler_type: A type of extra handler to add.
        :param extra_handler: A :py:class:`~.GlobalExtraHandlerType` to add to the component as an extra handler.
        :type extra_handler: :py:data:`~.ExtraHandlerFunction`
        :return: `None`
        """
        target = (
            self.before_handler if global_extra_handler_type is GlobalExtraHandlerType.BEFORE else self.after_handler
        )
        target.functions.append(extra_handler)

    def _get_runtime_info(self, ctx: Context) -> ServiceRuntimeInfo:
        """
        Method for retrieving runtime info about this component.

        :param ctx: Current dialog :py:class:`~.Context`.
        :return: :py:class:`~.chatsky.script.typing.ServiceRuntimeInfo`
            object where all not set fields are replaced with `[None]`.
        """
        return ServiceRuntimeInfo(
            name=self.name if self.name is not None else "[None]",
            path=self.path if self.path is not None else "[None]",
            timeout=self.timeout,
            asynchronous=self.asynchronous,
            execution_state=ctx.framework_data.service_states.copy(),
        )

    @property
    def info_dict(self) -> dict:
        """
        Property for retrieving info dictionary about this component.
        All not set fields there are replaced with `[None]`.

        :return: Info dict, containing most important component public fields as well as its type.
        """
        return {
            "type": type(self).__name__,
            "name": self.name,
            "path": self.path if self.path is not None else "[None]",
            "asynchronous": self.asynchronous,
            "start_condition": self.start_condition.__name__,
            "extra_handlers": {
                "before": self.before_handler.info_dict,
                "after": self.after_handler.info_dict,
            },
        }



"""
Mongo
-----
The Mongo module provides a MongoDB-based version of the :py:class:`.DBContextStorage` class.
This class is used to store and retrieve context data in a MongoDB.
It allows Chatsky to easily store and retrieve context data in a format that is highly scalable
and easy to work with.

MongoDB is a widely-used, open-source NoSQL database that is known for its scalability and performance.
It stores data in a format similar to JSON, making it easy to work with the data in a variety of programming languages
and environments. Additionally, MongoDB is highly scalable and can handle large amounts of data
and high levels of read and write traffic.
"""

from typing import Hashable, Dict, Any

try:
    from motor.motor_asyncio import AsyncIOMotorClient
    from bson.objectid import ObjectId

    mongo_available = True
except ImportError:
    mongo_available = False
    AsyncIOMotorClient = None
    ObjectId = Any

import json

from chatsky.script import Context

from .database import DBContextStorage, threadsafe_method
from .protocol import get_protocol_install_suggestion


class MongoContextStorage(DBContextStorage):
    """
    Implements :py:class:`.DBContextStorage` with `mongodb` as the database backend.

    :param path: Database URI. Example: `mongodb://user:password@host:port/dbname`.
    :param collection: Name of the collection to store the data in.
    """

    def __init__(self, path: str, collection: str = "context_collection"):
        DBContextStorage.__init__(self, path)
        if not mongo_available:
            install_suggestion = get_protocol_install_suggestion("mongodb")
            raise ImportError("`mongodb` package is missing.\n" + install_suggestion)
        self._mongo = AsyncIOMotorClient(self.full_path)
        db = self._mongo.get_default_database()
        self.collection = db[collection]

    @staticmethod
    def _adjust_key(key: Hashable) -> Dict[str, ObjectId]:
        """Convert a n-digit context id to a 24-digit mongo id"""
        new_key = hex(int.from_bytes(str.encode(str(key)), "big", signed=False))[3:]
        new_key = (new_key * (24 // len(new_key) + 1))[:24]
        assert len(new_key) == 24
        return {"_id": ObjectId(new_key)}

    @threadsafe_method
    async def set_item_async(self, key: Hashable, value: Context):
        new_key = self._adjust_key(key)
        value = value if isinstance(value, Context) else Context.cast(value)
        document = json.loads(value.model_dump_json())

        document.update(new_key)
        await self.collection.replace_one(new_key, document, upsert=True)

    @threadsafe_method
    async def get_item_async(self, key: Hashable) -> Context:
        adjust_key = self._adjust_key(key)
        document = await self.collection.find_one(adjust_key)
        if document:
            document.pop("_id")
            ctx = Context.cast(document)
            return ctx
        raise KeyError

    @threadsafe_method
    async def del_item_async(self, key: Hashable):
        adjust_key = self._adjust_key(key)
        await self.collection.delete_one(adjust_key)

    @threadsafe_method
    async def contains_async(self, key: Hashable) -> bool:
        adjust_key = self._adjust_key(key)
        return bool(await self.collection.find_one(adjust_key))

    @threadsafe_method
    async def len_async(self) -> int:
        return await self.collection.estimated_document_count()

    @threadsafe_method
    async def clear_async(self):
        await self.collection.delete_many(dict())



"""
Service Group
-------------
The Service Group module contains the
:py:class:`~.ServiceGroup` class, which is used to represent a group of related services.
This class provides a way to organize and manage multiple services as a single unit,
allowing for easier management and organization of the services within the pipeline.
The :py:class:`~.ServiceGroup` serves the important function of grouping services to work together in parallel.
"""

from __future__ import annotations
import asyncio
import logging
from typing import Optional, List, Union, Awaitable, TYPE_CHECKING

from chatsky.script import Context

from .utils import collect_defined_constructor_parameters_to_dict, _get_attrs_with_updates
from ..pipeline.component import PipelineComponent
from ..types import (
    StartConditionCheckerFunction,
    ComponentExecutionState,
    ServiceGroupBuilder,
    GlobalExtraHandlerType,
    ExtraHandlerConditionFunction,
    ExtraHandlerFunction,
    ExtraHandlerBuilder,
    ExtraHandlerType,
)
from .service import Service

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from chatsky.pipeline.pipeline.pipeline import Pipeline


class ServiceGroup(PipelineComponent):
    """
    A service group class.
    Service group can be included into pipeline as an object or a pipeline component list.
    Service group can be synchronous or asynchronous.
    Components in synchronous groups are executed consequently (no matter is they are synchronous or asynchronous).
    Components in asynchronous groups are executed simultaneously.
    Group can be asynchronous only if all components in it are asynchronous.

    :param components: A `ServiceGroupBuilder` object, that will be added to the group.
    :type components: :py:data:`~.ServiceGroupBuilder`
    :param before_handler: List of `ExtraHandlerBuilder` to add to the group.
    :type before_handler: Optional[:py:data:`~.ExtraHandlerBuilder`]
    :param after_handler: List of `ExtraHandlerBuilder` to add to the group.
    :type after_handler: Optional[:py:data:`~.ExtraHandlerBuilder`]
    :param timeout: Timeout to add to the group.
    :param asynchronous: Requested asynchronous property.
    :param start_condition: :py:data:`~.StartConditionCheckerFunction` that is invoked before each group execution;
        group is executed only if it returns `True`.
    :param name: Requested group name.
    """

    def __init__(
        self,
        components: ServiceGroupBuilder,
        before_handler: Optional[ExtraHandlerBuilder] = None,
        after_handler: Optional[ExtraHandlerBuilder] = None,
        timeout: Optional[float] = None,
        asynchronous: Optional[bool] = None,
        start_condition: Optional[StartConditionCheckerFunction] = None,
        name: Optional[str] = None,
    ):
        overridden_parameters = collect_defined_constructor_parameters_to_dict(
            before_handler=before_handler,
            after_handler=after_handler,
            timeout=timeout,
            asynchronous=asynchronous,
            start_condition=start_condition,
            name=name,
        )
        if isinstance(components, ServiceGroup):
            self.__init__(
                **_get_attrs_with_updates(
                    components,
                    (
                        "calculated_async_flag",
                        "path",
                    ),
                    {"requested_async_flag": "asynchronous"},
                    overridden_parameters,
                )
            )
        elif isinstance(components, dict):
            components.update(overridden_parameters)
            self.__init__(**components)
        elif isinstance(components, List):
            self.components = self._create_components(components)
            calc_async = all([service.asynchronous for service in self.components])
            super(ServiceGroup, self).__init__(
                before_handler, after_handler, timeout, asynchronous, calc_async, start_condition, name
            )
        else:
            raise Exception(f"Unknown type for ServiceGroup {components}")

    async def _run_services_group(self, ctx: Context, pipeline: Pipeline) -> None:
        """
        Method for running this service group.
        It doesn't include extra handlers execution, start condition checking or error handling - pure execution only.
        Executes components inside the group based on its `asynchronous` property.
        Collects information about their execution state - group is finished successfully
        only if all components in it finished successfully.

        :param ctx: Current dialog context.
        :param pipeline: The current pipeline.
        """
        self._set_state(ctx, ComponentExecutionState.RUNNING)

        if self.asynchronous:
            service_futures = [service(ctx, pipeline) for service in self.components]
            for service, future in zip(self.components, await asyncio.gather(*service_futures, return_exceptions=True)):
                service_result = future
                if service.asynchronous and isinstance(service_result, Awaitable):
                    await service_result
                elif isinstance(service_result, asyncio.TimeoutError):
                    logger.warning(f"{type(service).__name__} '{service.name}' timed out!")

        else:
            for service in self.components:
                service_result = await service(ctx, pipeline)
                if service.asynchronous and isinstance(service_result, Awaitable):
                    await service_result

        failed = any([service.get_state(ctx) == ComponentExecutionState.FAILED for service in self.components])
        self._set_state(ctx, ComponentExecutionState.FAILED if failed else ComponentExecutionState.FINISHED)

    async def _run(
        self,
        ctx: Context,
        pipeline: Pipeline,
    ) -> None:
        """
        Method for handling this group execution.
        Executes extra handlers before and after execution, checks start condition and catches runtime exceptions.

        :param ctx: Current dialog context.
        :param pipeline: The current pipeline.
        """
        await self.run_extra_handler(ExtraHandlerType.BEFORE, ctx, pipeline)

        try:
            if self.start_condition(ctx, pipeline):
                await self._run_services_group(ctx, pipeline)
            else:
                self._set_state(ctx, ComponentExecutionState.NOT_RUN)

        except Exception as exc:
            self._set_state(ctx, ComponentExecutionState.FAILED)
            logger.error(f"ServiceGroup '{self.name}' execution failed!", exc_info=exc)

        await self.run_extra_handler(ExtraHandlerType.AFTER, ctx, pipeline)

    def log_optimization_warnings(self):
        """
        Method for logging service group optimization warnings for all this groups inner components.
        (NOT this group itself!).
        Warnings are basically messages,
        that indicate service group inefficiency or explicitly defined parameters mismatch.
        These are cases for warnings issuing:

        - Service can be asynchronous, however is marked synchronous explicitly.
        - Service is not asynchronous, however has a timeout defined.
        - Group is not marked synchronous explicitly and contains both synchronous and asynchronous components.

        :return: `None`
        """
        for service in self.components:
            if isinstance(service, Service):
                if (
                    service.calculated_async_flag
                    and service.requested_async_flag is not None
                    and not service.requested_async_flag
                ):
                    logger.warning(f"Service '{service.name}' could be asynchronous!")
                if not service.asynchronous and service.timeout is not None:
                    logger.warning(f"Timeout can not be applied for Service '{service.name}': it's not asynchronous!")
            else:
                if not service.calculated_async_flag:
                    if service.requested_async_flag is None and any(
                        [sub_service.asynchronous for sub_service in service.components]
                    ):
                        logger.warning(
                            f"ServiceGroup '{service.name}' contains both sync and async services, "
                            "it should be split or marked as synchronous explicitly!",
                        )
                service.log_optimization_warnings()

    def add_extra_handler(
        self,
        global_extra_handler_type: GlobalExtraHandlerType,
        extra_handler: ExtraHandlerFunction,
        condition: ExtraHandlerConditionFunction = lambda _: True,
    ):
        """
        Method for adding a global extra handler to this group.
        Adds extra handler to itself and propagates it to all inner components.
        Uses a special condition function to determine whether to add extra handler to any particular inner component.
        Condition checks components path to be in whitelist (if defined) and not to be in blacklist (if defined).

        :param global_extra_handler_type: A type of extra handler to add.
        :param extra_handler: A `ExtraHandlerFunction` to add as an extra handler.
        :type extra_handler: :py:data:`~.ExtraHandlerFunction`
        :param condition: A condition function.
        :return: `None`
        """
        super().add_extra_handler(global_extra_handler_type, extra_handler)
        for service in self.components:
            if not condition(service.path):
                continue
            if isinstance(service, Service):
                service.add_extra_handler(global_extra_handler_type, extra_handler)
            else:
                service.add_extra_handler(global_extra_handler_type, extra_handler, condition)

    @property
    def info_dict(self) -> dict:
        """
        See `Component.info_dict` property.
        Adds `services` key to base info dictionary.
        """
        representation = super(ServiceGroup, self).info_dict
        representation.update({"services": [service.info_dict for service in self.components]})
        return representation

    @staticmethod
    def _create_components(services: ServiceGroupBuilder) -> List[Union[Service, "ServiceGroup"]]:
        """
        Utility method, used to create inner components, judging by their nature.
        Services are created from services and dictionaries.
        ServiceGroups are created from service groups and lists.

        :param services: ServiceGroupBuilder object (a `ServiceGroup` instance or a list).
        :type services: :py:data:`~.ServiceGroupBuilder`
        :return: List of services and service groups.
        """
        handled_services: List[Union[Service, "ServiceGroup"]] = []
        for service in services:
            if isinstance(service, List) or isinstance(service, ServiceGroup):
                handled_services.append(ServiceGroup(service))
            else:
                handled_services.append(Service(service))
        return handled_services



"""
Singleton Turn Caching
----------------------
This module contains functions for caching function results on each dialog turn.
"""

import functools
from typing import Callable, List, Optional


USED_CACHES: List[Callable] = list()
"""Cache singleton, it is common for all actors and pipelines in current environment."""


def cache_clear():
    """
    Function for cache singleton clearing, it is called in the end of pipeline execution turn.
    """
    for used_cache in USED_CACHES:
        used_cache.cache_clear()


def lru_cache(maxsize: Optional[int] = None, typed: bool = False) -> Callable:
    """
    Decorator function for caching function results in scripts.
    Works like the standard :py:func:`~functools.lru_cache` function.
    Caches are kept in a library-wide singleton and cleared in the end of each turn.
    """

    def decorator(func):
        global USED_CACHES

        @functools.wraps(func)
        @functools.lru_cache(maxsize=maxsize, typed=typed)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        USED_CACHES += [wrapper]
        return wrapper

    return decorator


def cache(func):
    """
    Decorator function for caching function results in scripts.
    Works like the standard :py:func:`~functools.cache` function.
    Caches are kept in a library-wide singleton and cleared in the end of each turn.
    """
    return lru_cache(maxsize=None)(func)



"""
Conditions
---------------------------
Provides slot-related conditions.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from chatsky.script import Context
    from chatsky.slots.slots import SlotName
    from chatsky.pipeline import Pipeline


def slots_extracted(*slots: SlotName, mode: Literal["any", "all"] = "all"):
    """
    Conditions that checks if slots are extracted.

    :param slots: Names for slots that need to be checked.
    :param mode: Whether to check if all slots are extracted or any slot is extracted.
    """

    def check_slot_state(ctx: Context, pipeline: Pipeline) -> bool:
        manager = ctx.framework_data.slot_manager
        if mode == "all":
            return all(manager.is_slot_extracted(slot) for slot in slots)
        elif mode == "any":
            return any(manager.is_slot_extracted(slot) for slot in slots)
        raise ValueError(f"{mode!r} not in ['any', 'all'].")

    return check_slot_state



"""
Cleanup DB
----------
This module defines functions that allow to delete data in various types of databases,
including JSON, MongoDB, Pickle, Redis, Shelve, SQL, and YDB databases.
"""

import os

from chatsky.context_storages import (
    JSONContextStorage,
    MongoContextStorage,
    PickleContextStorage,
    RedisContextStorage,
    ShelveContextStorage,
    SQLContextStorage,
    YDBContextStorage,
    json_available,
    mongo_available,
    pickle_available,
    redis_available,
    sqlite_available,
    postgres_available,
    mysql_available,
    ydb_available,
)


async def delete_json(storage: JSONContextStorage):
    """
    Delete all data from a JSON context storage.

    :param storage: A JSONContextStorage object.
    """
    if not json_available:
        raise Exception("Can't delete JSON database - JSON provider unavailable!")
    if os.path.isfile(storage.path):
        os.remove(storage.path)


async def delete_mongo(storage: MongoContextStorage):
    """
    Delete all data from a MongoDB context storage.

    :param storage: A MongoContextStorage object
    """
    if not mongo_available:
        raise Exception("Can't delete mongo database - mongo provider unavailable!")
    await storage.collection.drop()


async def delete_pickle(storage: PickleContextStorage):
    """
    Delete all data from a Pickle context storage.

    :param storage: A PickleContextStorage object.
    """
    if not pickle_available:
        raise Exception("Can't delete pickle database - pickle provider unavailable!")
    if os.path.isfile(storage.path):
        os.remove(storage.path)


async def delete_redis(storage: RedisContextStorage):
    """
    Delete all data from a Redis context storage.

    :param storage: A RedisContextStorage object.
    """
    if not redis_available:
        raise Exception("Can't delete redis database - redis provider unavailable!")
    await storage.clear_async()


async def delete_shelve(storage: ShelveContextStorage):
    """
    Delete all data from a Shelve context storage.

    :param storage: A ShelveContextStorage object.
    """
    if os.path.isfile(storage.path):
        os.remove(storage.path)


async def delete_sql(storage: SQLContextStorage):
    """
    Delete all data from an SQL context storage.

    :param storage: An SQLContextStorage object.
    """
    if storage.dialect == "postgres" and not postgres_available:
        raise Exception("Can't delete postgres database - postgres provider unavailable!")
    if storage.dialect == "sqlite" and not sqlite_available:
        raise Exception("Can't delete sqlite database - sqlite provider unavailable!")
    if storage.dialect == "mysql" and not mysql_available:
        raise Exception("Can't delete mysql database - mysql provider unavailable!")
    async with storage.engine.connect() as conn:
        await conn.run_sync(storage.table.drop, storage.engine)


async def delete_ydb(storage: YDBContextStorage):
    """
    Delete all data from a YDB context storage.

    :param storage: A YDBContextStorage object.
    """
    if not ydb_available:
        raise Exception("Can't delete ydb database - ydb provider unavailable!")

    async def callee(session):
        await session.drop_table("/".join([storage.database, storage.table_name]))

    await storage.pool.retry_operation(callee)



"""
Message
-------
The :py:class:`.Message` class is a universal data model for representing a message that should be supported by
Chatsky. It only contains types and properties that are compatible with most messaging services.
"""

from typing import Literal, Optional, List, Union
from pathlib import Path
from urllib.request import urlopen
import uuid
import abc

from pydantic import Field, FilePath, HttpUrl, model_validator
from pydantic_core import Url

from chatsky.messengers.common.interface import MessengerInterfaceWithAttachments
from chatsky.utils.devel import JSONSerializableDict, PickleEncodedValue, JSONSerializableExtras


class DataModel(JSONSerializableExtras):
    """
    This class is a Pydantic BaseModel that can have any type and number of extras.
    """

    pass


class Attachment(DataModel, abc.ABC):
    """
    Chatsky Message attachment base class.
    It is capable of serializing and validating all the model fields to JSON.
    """

    chatsky_attachment_type: str


class CallbackQuery(Attachment):
    """
    This class is a data model that represents a callback query attachment.
    It is sent as a response to non-message events, e.g. keyboard UI interactions.
    It has query string attribute, that represents the response data string.
    """

    query_string: Optional[str]
    chatsky_attachment_type: Literal["callback_query"] = "callback_query"


class Location(Attachment):
    """
    This class is a data model that represents a geographical
    location on the Earth's surface.
    It has two attributes, longitude and latitude, both of which are float values.
    If the absolute difference between the latitude and longitude values of the two
    locations is less than 0.00004, they are considered equal.
    """

    longitude: float
    latitude: float
    chatsky_attachment_type: Literal["location"] = "location"


class Contact(Attachment):
    """
    This class is a data model that represents a contact.
    It includes phone number, and user first and last name.
    """

    phone_number: str
    first_name: str
    last_name: Optional[str]
    chatsky_attachment_type: Literal["contact"] = "contact"


class Invoice(Attachment):
    """
    This class is a data model that represents an invoice.
    It includes title, description, currency name and amount.
    """

    title: str
    description: str
    currency: str
    amount: int
    chatsky_attachment_type: Literal["invoice"] = "invoice"


class PollOption(DataModel):
    """
    This class is a data model that represents a poll option.
    It includes the option name and votes number.
    """

    text: str
    votes: int = Field(default=0)
    chatsky_attachment_type: Literal["poll_option"] = "poll_option"


class Poll(Attachment):
    """
    This class is a data model that represents a poll.
    It includes a list of poll options.
    """

    question: str
    options: List[PollOption]
    chatsky_attachment_type: Literal["poll"] = "poll"


class DataAttachment(Attachment):
    """
    This class represents an attachment that can be either
    a local file, a URL to a file or a ID of a file on a certain server (such as telegram).
    This attachment can also be optionally cached for future use.
    """

    source: Optional[Union[HttpUrl, FilePath]] = None
    """Attachment source -- either a URL to a file or a local filepath."""
    use_cache: bool = True
    """
    Whether to cache the file (only for URL and ID files).
    Disable this if you want to always respond with the most up-to-date version of the file.
    """
    cached_filename: Optional[Path] = None
    """
    This field is used to store a path to cached version of this file (retrieved from id or URL).
    This field is managed by framework.
    """
    id: Optional[str] = None
    """
    ID of the file on a file server (e.g. file_id for telegram attachments).
    :py:meth:`~.MessengerInterfaceWithAttachments.get_attachment_bytes` is used to retrieve bytes from ID.
    """

    async def _cache_attachment(self, data: bytes, directory: Path) -> None:
        """
        Cache attachment, save bytes into a file.
        File has a UUID name based on its `self.source` or `self.id`.

        :param data: attachment data bytes.
        :param directory: cache directory where attachment will be saved.
        """

        filename = str(uuid.uuid5(uuid.NAMESPACE_URL, str(self.source or self.id)))
        self.cached_filename = directory / filename
        self.cached_filename.write_bytes(data)

    async def get_bytes(self, from_interface: MessengerInterfaceWithAttachments) -> Optional[bytes]:
        """
        Retrieve attachment bytes.
        If the attachment is represented by URL or saved in a file,
        it will be downloaded or read automatically.
        If cache use is allowed and the attachment is cached, cached file will be used.
        Otherwise, a :py:meth:`~.MessengerInterfaceWithAttachments.get_attachment_bytes`
        will be used for receiving attachment bytes via ID.

        If cache use is allowed and the attachment is a URL or an ID, bytes will be cached locally.

        :param from_interface: messenger interface the attachment was received from.
        """

        if isinstance(self.source, Path):
            with open(self.source, "rb") as file:
                return file.read()
        elif self.use_cache and self.cached_filename is not None and self.cached_filename.exists():
            with open(self.cached_filename, "rb") as file:
                return file.read()
        elif isinstance(self.source, Url):
            with urlopen(self.source.unicode_string()) as url:
                attachment_data = url.read()
        else:
            attachment_data = await from_interface.get_attachment_bytes(self.id)
        if self.use_cache:
            await self._cache_attachment(attachment_data, from_interface.attachments_directory)
        return attachment_data

    @model_validator(mode="before")
    @classmethod
    def validate_source_or_id(cls, values: dict):
        if not isinstance(values, dict):
            raise AssertionError(f"Invalid constructor parameters: {str(values)}")
        if bool(values.get("source")) == bool(values.get("id")):
            raise AssertionError("Attachment type requires exactly one parameter, `source` or `id`, to be set.")
        return values


class Audio(DataAttachment):
    """Represents an audio file attachment."""

    chatsky_attachment_type: Literal["audio"] = "audio"


class Video(DataAttachment):
    """Represents a video file attachment."""

    chatsky_attachment_type: Literal["video"] = "video"


class Animation(DataAttachment):
    """Represents an animation file attachment."""

    chatsky_attachment_type: Literal["animation"] = "animation"


class Image(DataAttachment):
    """Represents an image file attachment."""

    chatsky_attachment_type: Literal["image"] = "image"


class Sticker(DataAttachment):
    """Represents a sticker as a file attachment."""

    chatsky_attachment_type: Literal["sticker"] = "sticker"


class Document(DataAttachment):
    """Represents a document file attachment."""

    chatsky_attachment_type: Literal["document"] = "document"


class VoiceMessage(DataAttachment):
    """Represents a voice message."""

    chatsky_attachment_type: Literal["voice_message"] = "voice_message"


class VideoMessage(DataAttachment):
    """Represents a video message."""

    chatsky_attachment_type: Literal["video_message"] = "video_message"


class MediaGroup(Attachment):
    """
    Represents a group of media attachments.
    Without this class attachments are sent one-by-one.

    Be mindful of limitations that certain services apply
    (e.g. Telegram does not allow audio or document files to be mixed with other types when using media groups,
    so you should send them separately by putting them directly in :py:attr:`~.Message.attachments`).
    """

    group: List[Union[Audio, Video, Image, Document, DataAttachment]] = Field(default_factory=list)
    chatsky_attachment_type: Literal["media_group"] = "media_group"


class Message(DataModel):
    """
    Class representing a message and contains several
    class level variables to store message information.

    It includes message text, list of attachments, annotations,
    MISC dictionary (that consists of user-defined parameters)
    and original message field that represents
    the update received from messenger interface API.
    """

    text: Optional[str] = None
    attachments: Optional[
        List[
            Union[
                CallbackQuery,
                Location,
                Contact,
                Invoice,
                Poll,
                Audio,
                Video,
                Animation,
                Image,
                Sticker,
                Document,
                VoiceMessage,
                VideoMessage,
                MediaGroup,
                DataModel,
            ]
        ]
    ] = None
    annotations: Optional[JSONSerializableDict] = None
    misc: Optional[JSONSerializableDict] = None
    original_message: Optional[PickleEncodedValue] = None

    def __init__(
        self,
        text: Optional[str] = None,
        attachments: Optional[
            List[
                Union[
                    CallbackQuery,
                    Location,
                    Contact,
                    Invoice,
                    Poll,
                    Audio,
                    Video,
                    Animation,
                    Image,
                    Sticker,
                    Document,
                    VoiceMessage,
                    VideoMessage,
                    MediaGroup,
                ]
            ]
        ] = None,
        annotations: Optional[JSONSerializableDict] = None,
        misc: Optional[JSONSerializableDict] = None,
        **kwargs,
    ):
        super().__init__(text=text, attachments=attachments, annotations=annotations, misc=misc, **kwargs)

    def __repr__(self) -> str:
        return " ".join([f"{key}='{value}'" for key, value in self.model_dump(exclude_none=True).items()])



"""
Database
--------
The `Database` module provides classes for managing the context storage of a dialog system.
The module can be used to store information such as the current state of the conversation
and other data. This module includes the intermediate class (:py:class:`.DBContextStorage`) is a class
that developers can inherit from in order to create their own context storage solutions.
This class implements the basic functionality and can be extended to add additional features as needed.
"""

import asyncio
import importlib
import threading
from functools import wraps
from abc import ABC, abstractmethod
from typing import Callable, Hashable, Optional

from .protocol import PROTOCOLS
from ..script import Context


class DBContextStorage(ABC):
    r"""
    An abstract interface for `chatsky` DB context storages.
    It includes the most essential methods of the python `dict` class.
    Can not be instantiated.

    :param path: Parameter `path` should be set with the URI of the database.
        It includes a prefix and the required connection credentials.
        Example: postgresql+asyncpg://user:password@host:port/database
        In the case of classes that save data to hard drive instead of external databases
        you need to specify the location of the file, like you do in sqlite.
        Keep in mind that in Windows you will have to use double backslashes '\\'
        instead of forward slashes '/' when defining the file path.

    """

    def __init__(self, path: str):
        _, _, file_path = path.partition("://")
        self.full_path = path
        """Full path to access the context storage, as it was provided by user."""
        self.path = file_path
        """`full_path` without a prefix defining db used"""
        self._lock = threading.Lock()
        """Threading for methods that require single thread access."""

    def __getitem__(self, key: Hashable) -> Context:
        """
        Synchronous method for accessing stored Context.

        :param key: Hashable key used to store Context instance.
        :return: The stored context, associated with the given key.
        """
        return asyncio.run(self.get_item_async(key))

    @abstractmethod
    async def get_item_async(self, key: Hashable) -> Context:
        """
        Asynchronous method for accessing stored Context.

        :param key: Hashable key used to store Context instance.
        :return: The stored context, associated with the given key.
        """
        raise NotImplementedError

    def __setitem__(self, key: Hashable, value: Context):
        """
        Synchronous method for storing Context.

        :param key: Hashable key used to store Context instance.
        :param value: Context to store.
        """
        return asyncio.run(self.set_item_async(key, value))

    @abstractmethod
    async def set_item_async(self, key: Hashable, value: Context):
        """
        Asynchronous method for storing Context.

        :param key: Hashable key used to store Context instance.
        :param value: Context to store.
        """
        raise NotImplementedError

    def __delitem__(self, key: Hashable):
        """
        Synchronous method for removing stored Context.

        :param key: Hashable key used to identify Context instance for deletion.
        """
        return asyncio.run(self.del_item_async(key))

    @abstractmethod
    async def del_item_async(self, key: Hashable):
        """
        Asynchronous method for removing stored Context.

        :param key: Hashable key used to identify Context instance for deletion.
        """
        raise NotImplementedError

    def __contains__(self, key: Hashable) -> bool:
        """
        Synchronous method for finding whether any Context is stored with given key.

        :param key: Hashable key used to check if Context instance is stored.
        :return: True if there is Context accessible by given key, False otherwise.
        """
        return asyncio.run(self.contains_async(key))

    @abstractmethod
    async def contains_async(self, key: Hashable) -> bool:
        """
        Asynchronous method for finding whether any Context is stored with given key.

        :param key: Hashable key used to check if Context instance is stored.
        :return: True if there is Context accessible by given key, False otherwise.
        """
        raise NotImplementedError

    def __len__(self) -> int:
        """
        Synchronous method for retrieving number of stored Contexts.

        :return: The number of stored Contexts.
        """
        return asyncio.run(self.len_async())

    @abstractmethod
    async def len_async(self) -> int:
        """
        Asynchronous method for retrieving number of stored Contexts.

        :return: The number of stored Contexts.
        """
        raise NotImplementedError

    def get(self, key: Hashable, default: Optional[Context] = None) -> Context:
        """
        Synchronous method for accessing stored Context, returning default if no Context is stored with the given key.

        :param key: Hashable key used to store Context instance.
        :param default: Optional default value to be returned if no Context is found.
        :return: The stored context, associated with the given key or default value.
        """
        return asyncio.run(self.get_async(key, default))

    async def get_async(self, key: Hashable, default: Optional[Context] = None) -> Context:
        """
        Asynchronous method for accessing stored Context, returning default if no Context is stored with the given key.

        :param key: Hashable key used to store Context instance.
        :param default: Optional default value to be returned if no Context is found.
        :return: The stored context, associated with the given key or default value.
        """
        try:
            return await self.get_item_async(str(key))
        except KeyError:
            return default

    def clear(self):
        """
        Synchronous method for clearing context storage, removing all the stored Contexts.
        """
        return asyncio.run(self.clear_async())

    @abstractmethod
    async def clear_async(self):
        """
        Asynchronous method for clearing context storage, removing all the stored Contexts.
        """
        raise NotImplementedError


def threadsafe_method(func: Callable):
    """
    A decorator that makes sure methods of an object instance are threadsafe.
    """

    @wraps(func)
    def _synchronized(self, *args, **kwargs):
        with self._lock:
            return func(self, *args, **kwargs)

    return _synchronized


def context_storage_factory(path: str, **kwargs) -> DBContextStorage:
    """
    Use context_storage_factory to lazy import context storage types and instantiate them.
    The function takes a database connection URI or its equivalent. It should be prefixed with database name,
    followed by the symbol triplet '://'.

    Then, you should list the connection parameters like this: user:password@host:port/database
    The whole URI will then look like this:

    - shelve://path_to_the_file/file_name
    - json://path_to_the_file/file_name
    - pickle://path_to_the_file/file_name
    - sqlite+aiosqlite://path_to_the_file/file_name
    - redis://:pass@localhost:6378/0
    - mongodb://admin:pass@localhost:27016/admin
    - mysql+asyncmy://root:pass@localhost:3306/test
    - postgresql+asyncpg://postgres:pass@localhost:5430/test
    - grpc://localhost:2134/local
    - grpcs://localhost:2134/local

    For context storages that write to local files, the function expects a file path instead of connection params:
    json://file.json
    When using sqlite backend your prefix should contain three slashes if you use Windows, or four in other cases:
    sqlite:////file.db
    If you want to use additional parameters in class constructors, you can pass them to this function as kwargs.

    :param path: Path to the file.
    """
    prefix, _, _ = path.partition("://")
    if "sql" in prefix:
        prefix = prefix.split("+")[0]  # this takes care of alternative sql drivers
    assert (
        prefix in PROTOCOLS
    ), f"""
    URI path should be prefixed with one of the following:\n
    {", ".join(PROTOCOLS.keys())}.\n
    For more information, see the function doc:\n{context_storage_factory.__doc__}
    """
    _class, module = PROTOCOLS[prefix]["class"], PROTOCOLS[prefix]["module"]
    target_class = getattr(importlib.import_module(f".{module}", package="chatsky.context_storages"), _class)
    return target_class(path, **kwargs)



"""
Slots
-----
This module defines base classes for slots and some concrete implementations of them.
"""

from __future__ import annotations

import asyncio
import re
from abc import ABC, abstractmethod
from typing import Callable, Any, Awaitable, TYPE_CHECKING, Union
from typing_extensions import TypeAlias, Annotated
import logging
from functools import reduce

from pydantic import BaseModel, model_validator, Field

from chatsky.utils.devel.async_helpers import wrap_sync_function_in_async
from chatsky.utils.devel.json_serialization import PickleEncodedValue

if TYPE_CHECKING:
    from chatsky.script import Context, Message
    from chatsky.pipeline.pipeline.pipeline import Pipeline


logger = logging.getLogger(__name__)


SlotName: TypeAlias = str
"""
A string to identify slots.

Top-level slots are identified by their key in a :py:class:`~.GroupSlot`.

E.g.

.. code:: python

    GroupSlot(
        user=RegexpSlot(),
        password=FunctionSlot,
    )

Has two slots with names "user" and "password".

For nested group slots use dots to separate names:

.. code:: python

    GroupSlot(
        user=GroupSlot(
            name=FunctionSlot,
            password=FunctionSlot,
        )
    )

Has two slots with names "user.name" and "user.password".
"""


def recursive_getattr(obj, slot_name: SlotName):
    def two_arg_getattr(__o, name):
        # pydantic handles exception when accessing a non-existing extra-field on its own
        # return None by default to avoid that
        return getattr(__o, name, None)

    return reduce(two_arg_getattr, [obj, *slot_name.split(".")])


def recursive_setattr(obj, slot_name: SlotName, value):
    parent_slot, _, slot = slot_name.rpartition(".")

    if parent_slot:
        setattr(recursive_getattr(obj, parent_slot), slot, value)
    else:
        setattr(obj, slot, value)


class SlotNotExtracted(Exception):
    """This exception can be returned or raised by slot extractor if slot extraction is unsuccessful."""

    pass


class ExtractedSlot(BaseModel, ABC):
    """
    Represents value of an extracted slot.

    Instances of this class are managed by framework and
    are stored in :py:attr:`~chatsky.script.core.context.FrameworkData.slot_manager`.
    They can be accessed via the ``ctx.framework_data.slot_manager.get_extracted_slot`` method.
    """

    @property
    @abstractmethod
    def __slot_extracted__(self) -> bool:
        """Whether the slot is extracted."""
        raise NotImplementedError

    def __unset__(self):
        """Mark slot as not extracted and clear extracted data (except for default value)."""
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        """String representation is used to fill templates."""
        raise NotImplementedError


class ExtractedValueSlot(ExtractedSlot):
    """Value extracted from :py:class:`~.ValueSlot`."""

    is_slot_extracted: bool
    extracted_value: PickleEncodedValue
    default_value: PickleEncodedValue = None

    @property
    def __slot_extracted__(self) -> bool:
        return self.is_slot_extracted

    def __unset__(self):
        self.is_slot_extracted = False
        self.extracted_value = SlotNotExtracted("Slot manually unset.")

    @property
    def value(self):
        """Extracted value or the default value if the slot is not extracted."""
        return self.extracted_value if self.is_slot_extracted else self.default_value

    def __str__(self):
        return str(self.value)


class ExtractedGroupSlot(ExtractedSlot, extra="allow"):
    __pydantic_extra__: dict[str, Union["ExtractedValueSlot", "ExtractedGroupSlot"]]

    @property
    def __slot_extracted__(self) -> bool:
        return all([slot.__slot_extracted__ for slot in self.__pydantic_extra__.values()])

    def __unset__(self):
        for child in self.__pydantic_extra__.values():
            child.__unset__()

    def __str__(self):
        return str({key: str(value) for key, value in self.__pydantic_extra__.items()})

    def update(self, old: "ExtractedGroupSlot"):
        """
        Rebase this extracted groups slot on top of another one.
        This is required to merge slot storage in-context
        with a potentially different slot configuration passed to pipeline.

        :param old: An instance of :py:class:`~.ExtractedGroupSlot` stored in-context.
            Extracted values will be transferred to this object.
        """
        for slot in old.__pydantic_extra__:
            if slot in self.__pydantic_extra__:
                new_slot = self.__pydantic_extra__[slot]
                old_slot = old.__pydantic_extra__[slot]
                if isinstance(new_slot, ExtractedGroupSlot) and isinstance(old_slot, ExtractedGroupSlot):
                    new_slot.update(old_slot)
                if isinstance(new_slot, ExtractedValueSlot) and isinstance(old_slot, ExtractedValueSlot):
                    self.__pydantic_extra__[slot] = old_slot


class BaseSlot(BaseModel, frozen=True):
    """
    BaseSlot is a base class for all slots.
    """

    @abstractmethod
    async def get_value(self, ctx: Context, pipeline: Pipeline) -> ExtractedSlot:
        """
        Extract slot value from :py:class:`~.Context` and return an instance of :py:class:`~.ExtractedSlot`.
        """
        raise NotImplementedError

    @abstractmethod
    def init_value(self) -> ExtractedSlot:
        """
        Provide an initial value to fill slot storage with.
        """
        raise NotImplementedError


class ValueSlot(BaseSlot, frozen=True):
    """
    Value slot is a base class for all slots that are designed to extract concrete values.
    Subclass it, if you want to declare your own slot type.
    """

    default_value: Any = None

    @abstractmethod
    async def extract_value(self, ctx: Context, pipeline: Pipeline) -> Union[Any, SlotNotExtracted]:
        """
        Return value extracted from context.

        Return :py:exc:`~.SlotNotExtracted` to mark extraction as unsuccessful.

        Raising exceptions is also allowed and will result in an unsuccessful extraction as well.
        """
        raise NotImplementedError

    async def get_value(self, ctx: Context, pipeline: Pipeline) -> ExtractedValueSlot:
        """Wrapper for :py:meth:`~.ValueSlot.extract_value` to handle exceptions."""
        extracted_value = SlotNotExtracted("Caught an exit exception.")
        is_slot_extracted = False

        try:
            extracted_value = await self.extract_value(ctx, pipeline)
            is_slot_extracted = not isinstance(extracted_value, SlotNotExtracted)
        except Exception as error:
            logger.exception(f"Exception occurred during {self.__class__.__name__!r} extraction.", exc_info=error)
            extracted_value = error
        finally:
            return ExtractedValueSlot.model_construct(
                is_slot_extracted=is_slot_extracted,
                extracted_value=extracted_value,
                default_value=self.default_value,
            )

    def init_value(self) -> ExtractedValueSlot:
        return ExtractedValueSlot.model_construct(
            is_slot_extracted=False,
            extracted_value=SlotNotExtracted("Initial slot extraction."),
            default_value=self.default_value,
        )


class GroupSlot(BaseSlot, extra="allow", frozen=True):
    """
    Base class for :py:class:`~.RootSlot` and :py:class:`~.GroupSlot`.
    """

    __pydantic_extra__: dict[str, Annotated[Union["GroupSlot", "ValueSlot"], Field(union_mode="left_to_right")]] = {}

    def __init__(self, **kwargs):  # supress unexpected argument warnings
        super().__init__(**kwargs)

    @model_validator(mode="after")
    def __check_extra_field_names__(self):
        """
        Extra field names cannot be dunder names or contain dots.
        """
        for field in self.__pydantic_extra__.keys():
            if "." in field:
                raise ValueError(f"Extra field name cannot contain dots: {field!r}")
            if field.startswith("__") and field.endswith("__"):
                raise ValueError(f"Extra field names cannot be dunder: {field!r}")
        return self

    async def get_value(self, ctx: Context, pipeline: Pipeline) -> ExtractedGroupSlot:
        child_values = await asyncio.gather(
            *(child.get_value(ctx, pipeline) for child in self.__pydantic_extra__.values())
        )
        return ExtractedGroupSlot(
            **{child_name: child_value for child_value, child_name in zip(child_values, self.__pydantic_extra__.keys())}
        )

    def init_value(self) -> ExtractedGroupSlot:
        return ExtractedGroupSlot(
            **{child_name: child.init_value() for child_name, child in self.__pydantic_extra__.items()}
        )


class RegexpSlot(ValueSlot, frozen=True):
    """
    RegexpSlot is a slot type that extracts its value using a regular expression.
    You can pass a compiled or a non-compiled pattern to the `regexp` argument.
    If you want to extract a particular group, but not the full match,
    change the `match_group_idx` parameter.
    """

    regexp: str
    match_group_idx: int = 0
    "Index of the group to match."

    async def extract_value(self, ctx: Context, _: Pipeline) -> Union[str, SlotNotExtracted]:
        request_text = ctx.last_request.text
        search = re.search(self.regexp, request_text)
        return (
            search.group(self.match_group_idx)
            if search
            else SlotNotExtracted(f"Failed to match pattern {self.regexp!r} in {request_text!r}.")
        )


class FunctionSlot(ValueSlot, frozen=True):
    """
    A simpler version of :py:class:`~.ValueSlot`.

    Uses a user-defined `func` to extract slot value from the :py:attr:`~.Context.last_request` Message.
    """

    func: Callable[[Message], Union[Awaitable[Union[Any, SlotNotExtracted]], Any, SlotNotExtracted]]

    async def extract_value(self, ctx: Context, _: Pipeline) -> Union[Any, SlotNotExtracted]:
        return await wrap_sync_function_in_async(self.func, ctx.last_request)


class SlotManager(BaseModel):
    """
    Provides API for managing slots.

    An instance of this class can be accessed via ``ctx.framework_data.slot_manager``.
    """

    slot_storage: ExtractedGroupSlot = Field(default_factory=ExtractedGroupSlot)
    """Slot storage. Stored inside ctx.framework_data."""
    root_slot: GroupSlot = Field(default_factory=GroupSlot, exclude=True)
    """Slot configuration passed during pipeline initialization."""

    def set_root_slot(self, root_slot: GroupSlot):
        """
        Set root_slot configuration from pipeline.
        Update extracted slots with the new configuration:

        New slots are added with their :py:meth:`~.BaseSlot.init_value`.
        Old extracted slot values are preserved only if their configuration did not change.
        That is if they are still present in the config and if their fundamental type did not change
        (i.e. `GroupSlot` did not turn into a `ValueSlot` or vice versa).

        This method is called by pipeline and is not supposed to be used otherwise.
        """
        self.root_slot = root_slot
        new_slot_storage = root_slot.init_value()
        new_slot_storage.update(self.slot_storage)
        self.slot_storage = new_slot_storage

    def get_slot(self, slot_name: SlotName) -> BaseSlot:
        """
        Get slot configuration from the slot name.

        :raises KeyError: If the slot with the specified name does not exist.
        """
        try:
            slot = recursive_getattr(self.root_slot, slot_name)
            if isinstance(slot, BaseSlot):
                return slot
        except (AttributeError, KeyError):
            pass
        raise KeyError(f"Could not find slot {slot_name!r}.")

    async def extract_slot(self, slot_name: SlotName, ctx: Context, pipeline: Pipeline) -> None:
        """
        Extract slot `slot_name` and store extracted value in `slot_storage`.

        :raises KeyError: If the slot with the specified name does not exist.
        """
        slot = self.get_slot(slot_name)
        value = await slot.get_value(ctx, pipeline)

        recursive_setattr(self.slot_storage, slot_name, value)

    async def extract_all(self, ctx: Context, pipeline: Pipeline):
        """
        Extract all slots from slot configuration `root_slot` and set `slot_storage` to the extracted value.
        """
        self.slot_storage = await self.root_slot.get_value(ctx, pipeline)

    def get_extracted_slot(self, slot_name: SlotName) -> ExtractedSlot:
        """
        Retrieve extracted value from `slot_storage`.

        :raises KeyError: If the slot with the specified name does not exist.
        """
        try:
            slot = recursive_getattr(self.slot_storage, slot_name)
            if isinstance(slot, ExtractedSlot):
                return slot
        except (AttributeError, KeyError):
            pass
        raise KeyError(f"Could not find slot {slot_name!r}.")

    def is_slot_extracted(self, slot_name: str) -> bool:
        """
        Return if the specified slot is extracted.

        :raises KeyError: If the slot with the specified name does not exist.
        """
        return self.get_extracted_slot(slot_name).__slot_extracted__

    def all_slots_extracted(self) -> bool:
        """
        Return if all slots are extracted.
        """
        return self.slot_storage.__slot_extracted__

    def unset_slot(self, slot_name: SlotName) -> None:
        """
        Mark specified slot as not extracted and clear extracted value.

        :raises KeyError: If the slot with the specified name does not exist.
        """
        self.get_extracted_slot(slot_name).__unset__()

    def unset_all_slots(self) -> None:
        """
        Mark all slots as not extracted and clear all extracted values.
        """
        self.slot_storage.__unset__()

    def fill_template(self, template: str) -> str:
        """
        Fill `template` string with extracted slot values and return a formatted string.

        `template` should be a format-string:

        E.g. "Your username is {profile.username}".

        For the example above, if ``profile.username`` slot has value "admin",
        it would return the following text:
        "Your username is admin".
        """
        return template.format(**dict(self.slot_storage.__pydantic_extra__.items()))



from typing import Any, Hashable, List, Optional, TextIO, Tuple
from uuid import uuid4
from chatsky.messengers.common.interface import PollingMessengerInterface
from chatsky.pipeline.types import PipelineRunnerFunction
from chatsky.script.core.context import Context
from chatsky.script.core.message import Message


class CLIMessengerInterface(PollingMessengerInterface):
    """
    Command line message interface is the default message interface, communicating with user via `STDIN/STDOUT`.
    This message interface can maintain dialog with one user at a time only.
    """

    supported_request_attachment_types = set()
    supported_response_attachment_types = set()

    def __init__(
        self,
        intro: Optional[str] = None,
        prompt_request: str = "request: ",
        prompt_response: str = "response: ",
        out_descriptor: Optional[TextIO] = None,
    ):
        super().__init__()
        self._ctx_id: Optional[Hashable] = None
        self._intro: Optional[str] = intro
        self._prompt_request: str = prompt_request
        self._prompt_response: str = prompt_response
        self._descriptor: Optional[TextIO] = out_descriptor

    def _request(self) -> List[Tuple[Message, Any]]:
        return [(Message(input(self._prompt_request)), self._ctx_id)]

    def _respond(self, responses: List[Context]):
        print(f"{self._prompt_response}{responses[0].last_response.text}", file=self._descriptor)

    async def connect(self, pipeline_runner: PipelineRunnerFunction, **kwargs):
        """
        The CLIProvider generates new dialog id used to user identification on each `connect` call.

        :param pipeline_runner: A function that should process user request and return context;
            usually it's a :py:meth:`~chatsky.pipeline.pipeline.pipeline.Pipeline._run_pipeline` function.
        :param \\**kwargs: argument, added for compatibility with super class, it shouldn't be used normally.
        """
        self._ctx_id = uuid4()
        if self._intro is not None:
            print(self._intro)
        await super().connect(pipeline_runner, **kwargs)



"""
Main
----
This module includes command line scripts for Superset dashboard configuration,
e.g. for creating and importing configuration archives.
In a configuration archive, you can define such settings as passwords, networking addressses etc.
using your own parameters that can be passed as a config file and overridden by command line arguments.

Examples
********

.. code:: bash

        # Create and import a configuration archive.
        # The import overrides existing dashboard configurations.
        chatsky.stats config.yaml \\
            -U superset_user \\
            -P superset_password \\
            -dP database_password \\
            --db.user=database_user \\
            --db.host=clickhouse \\
            --db.port=8123 \\
            --db.name=test \\
            --db.table=otel_logs \\
            --outfile=config_artifact.zip

"""

import sys
import argparse
from typing import Optional

from .cli import import_dashboard, make_zip_config
from .utils import PasswordAction


def main(parsed_args: Optional[argparse.Namespace] = None):
    """
    Function that evokes procedures defined in `cli` module.

    :param parsed_args: Set of command line arguments. If passed, overrides the command line contents.
        See the module docs for reference.
    """
    parser = argparse.ArgumentParser(
        usage="""Creates a configuration archive and uploads it to the Superset server.
        The import overrides existing dashboard configurations if present.
        The function accepts a yaml file; also, all of the options can also be overridden
        via the command line. Setting passwords interactively is supported.

        chatsky.stats config.yaml \\
            -U superset_user \\
            -P superset_password \\
            -dP database_password \\
            --db.user=database_user \\
            --db.host=clickhouse \\
            --db.port=8123 \\
            --db.name=test \\
            --db.table=otel_logs \\
            --outfile=config_artifact.zip

        Use the `--help` flag to get more information."""
    )
    parser.add_argument("file", type=str)
    parser.add_argument(
        "-dD",
        "--db.driver",
        choices=["clickhousedb+connect"],
        help="DBMS driver.",
        default="clickhousedb+connect",
    )
    parser.add_argument("-dU", "--db.user", help="Database user.")
    parser.add_argument("-dh", "--db.host", default="clickhouse", help="Database host.")
    parser.add_argument("-dp", "--db.port", help="Database port.")
    parser.add_argument("-dn", "--db.name", help="Name of the database.")
    parser.add_argument("-dt", "--db.table", default="otel_logs", help="Name of the table.")
    parser.add_argument("-o", "--outfile", help="Optionally persist the configuration as a zip file.")
    parser.add_argument("-i", "--infile", help="Configuration zip file to import.")
    parser.add_argument("-H", "--host", default="localhost", help="Superset host")
    parser.add_argument("-p", "--port", default="8088", help="Superset port.")
    parser.add_argument("-U", "--username", required=True, help="Superset user.")
    parser.add_argument(
        "-P",
        "--password",
        dest="password",
        type=str,
        action=PasswordAction,
        help="Superset password.",
        nargs="?",
        required=True,
    )
    parser.add_argument(
        "-dP",
        "--db.password",
        dest="db.password",
        type=str,
        action=PasswordAction,
        help="Database password.",
        required=True,
        nargs="?",
    )

    if parsed_args is None:
        parsed_args = parser.parse_args(sys.argv[1:])

    file = None
    use_infile = hasattr(parsed_args, "infile") and parsed_args.infile is not None
    use_outfile = hasattr(parsed_args, "outfile") and parsed_args.outfile is not None
    if not use_infile:
        file = make_zip_config(parsed_args)
    else:
        file = parsed_args.infile

    import_dashboard(parsed_args, zip_file=str(file))

    if not use_infile and not use_outfile:
        file.unlink()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter
from opentelemetry.sdk.trace.export import ConsoleSpanExporter
from opentelemetry.sdk._logs.export import InMemoryLogExporter, ConsoleLogExporter
from opentelemetry.sdk.metrics.export import InMemoryMetricReader, ConsoleMetricExporter
from .utils import get_extra_handler_name, set_logger_destination, set_tracer_destination
from .instrumentor import OtelInstrumentor, OTLPMetricExporter, OTLPLogExporter, OTLPSpanExporter



"""
Telegram Base
-------------
This module implements a base interface for interactions with the
Telegram API.
"""

from pathlib import Path
from typing import Any, Callable, Optional

from chatsky.utils.devel.extra_field_helpers import grab_extra_fields

from chatsky.messengers.common import MessengerInterfaceWithAttachments
from chatsky.pipeline.types import PipelineRunnerFunction
from chatsky.script.core.message import (
    Animation,
    Audio,
    CallbackQuery,
    Contact,
    Document,
    Image,
    Invoice,
    Location,
    Message,
    Poll,
    PollOption,
    Sticker,
    Video,
    VideoMessage,
    VoiceMessage,
    MediaGroup,
)

try:
    from telegram import (
        InputMediaAnimation,
        InputMediaAudio,
        InputMediaDocument,
        InputMediaPhoto,
        InputMediaVideo,
        Update,
        Message as TelegramMessage,
    )
    from telegram.ext import Application, ExtBot, MessageHandler, CallbackQueryHandler
    from telegram.ext.filters import ALL

    telegram_available = True
except ImportError:
    ExtBot = Any
    Update = Any
    TelegramMessage = Any

    telegram_available = False


class _AbstractTelegramInterface(MessengerInterfaceWithAttachments):
    """
    Messenger interface mixin for Telegram API usage.
    """

    supported_request_attachment_types = {
        Location,
        Contact,
        Poll,
        Sticker,
        Audio,
        Video,
        Animation,
        Image,
        Document,
        VoiceMessage,
        VideoMessage,
        Invoice,
    }
    supported_response_attachment_types = {
        Location,
        Contact,
        Poll,
        Sticker,
        Audio,
        Video,
        Animation,
        Image,
        Document,
        VoiceMessage,
        VideoMessage,
        MediaGroup,
    }

    def __init__(self, token: str, attachments_directory: Optional[Path] = None) -> None:
        super().__init__(attachments_directory)
        if not telegram_available:
            raise ImportError("`python-telegram-bot` package is missing.\nTry to run `pip install chatsky[telegram]`.")

        self.application = Application.builder().token(token).build()
        self.application.add_handler(MessageHandler(ALL, self.on_message))
        self.application.add_handler(CallbackQueryHandler(self.on_callback))

    async def get_attachment_bytes(self, source: str) -> bytes:
        file = await self.application.bot.get_file(source)
        data = await file.download_as_bytearray()
        return bytes(data)

    def extract_message_from_telegram(self, update: TelegramMessage) -> Message:
        """
        Convert Telegram update to Chatsky message.
        Extract text and supported attachments.

        :param update: Telegram update object.
        :return: Chatsky message object.
        """

        message = Message()
        message.attachments = list()

        message.text = update.text or update.caption
        if update.location is not None:
            message.attachments += [Location(latitude=update.location.latitude, longitude=update.location.longitude)]
        if update.contact is not None:
            message.attachments += [
                Contact(
                    phone_number=update.contact.phone_number,
                    first_name=update.contact.first_name,
                    last_name=update.contact.last_name,
                    user_id=update.contact.user_id,
                )
            ]
        if update.invoice is not None:
            message.attachments += [
                Invoice(
                    title=update.invoice.title,
                    description=update.invoice.description,
                    currency=update.invoice.currency,
                    amount=update.invoice.total_amount,
                )
            ]
        if update.poll is not None:
            message.attachments += [
                Poll(
                    question=update.poll.question,
                    options=[PollOption(text=option.text, votes=option.voter_count) for option in update.poll.options],
                    is_closed=update.poll.is_closed,
                    is_anonymous=update.poll.is_anonymous,
                    type=update.poll.type,
                    multiple_answers=update.poll.allows_multiple_answers,
                    correct_option_id=update.poll.correct_option_id,
                    explanation=update.poll.explanation,
                    open_period=update.poll.open_period,
                )
            ]
        if update.sticker is not None:
            message.attachments += [
                Sticker(
                    id=update.sticker.file_id,
                    is_animated=update.sticker.is_animated,
                    is_video=update.sticker.is_video,
                    type=update.sticker.type,
                )
            ]
        if update.audio is not None:
            thumbnail = (
                Image(id=update.audio.thumbnail.file_id, file_unique_id=update.audio.thumbnail.file_unique_id)
                if update.audio.thumbnail is not None
                else None
            )
            message.attachments += [
                Audio(
                    id=update.audio.file_id,
                    file_unique_id=update.audio.file_unique_id,
                    duration=update.audio.duration,
                    performer=update.audio.performer,
                    file_name=update.audio.file_name,
                    mime_type=update.audio.mime_type,
                    thumbnail=thumbnail,
                )
            ]
        if update.video is not None:
            thumbnail = (
                Image(id=update.video.thumbnail.file_id, file_unique_id=update.video.thumbnail.file_unique_id)
                if update.video.thumbnail is not None
                else None
            )
            message.attachments += [
                Video(
                    id=update.video.file_id,
                    file_unique_id=update.video.file_unique_id,
                    width=update.video.width,
                    height=update.video.height,
                    duration=update.video.duration,
                    file_name=update.video.file_name,
                    mime_type=update.video.mime_type,
                    thumbnail=thumbnail,
                )
            ]
        if update.animation is not None:
            thumbnail = (
                Image(id=update.animation.thumbnail.file_id, file_unique_id=update.animation.thumbnail.file_unique_id)
                if update.animation.thumbnail is not None
                else None
            )
            message.attachments += [
                Animation(
                    id=update.animation.file_id,
                    file_unique_id=update.animation.file_unique_id,
                    width=update.animation.width,
                    height=update.animation.height,
                    duration=update.animation.duration,
                    file_name=update.animation.file_name,
                    mime_type=update.animation.mime_type,
                    thumbnail=thumbnail,
                )
            ]
        if len(update.photo) > 0:
            message.attachments += [
                Image(
                    id=picture.file_id,
                    file_unique_id=picture.file_unique_id,
                    width=picture.width,
                    height=picture.height,
                )
                for picture in update.photo
            ]
        if update.document is not None:
            thumbnail = (
                Image(id=update.document.thumbnail.file_id, file_unique_id=update.document.thumbnail.file_unique_id)
                if update.document.thumbnail is not None
                else None
            )
            message.attachments += [
                Document(
                    id=update.document.file_id,
                    file_unique_id=update.document.file_unique_id,
                    file_name=update.document.file_name,
                    mime_type=update.document.mime_type,
                    thumbnail=thumbnail,
                )
            ]
        if update.voice is not None:
            message.attachments += [
                VoiceMessage(
                    id=update.voice.file_id,
                    file_unique_id=update.voice.file_unique_id,
                    mime_type=update.voice.mime_type,
                )
            ]
        if update.video_note is not None:
            thumbnail = (
                Image(id=update.video_note.thumbnail.file_id, file_unique_id=update.video_note.thumbnail.file_unique_id)
                if update.video_note.thumbnail is not None
                else None
            )
            message.attachments += [
                VideoMessage(
                    id=update.video_note.file_id,
                    file_unique_id=update.video_note.file_unique_id,
                    thumbnail=thumbnail,
                )
            ]

        return message

    async def cast_message_to_telegram_and_send(self, bot: ExtBot, chat_id: int, message: Message) -> None:
        """
        Send Chatsky message to Telegram.
        Sometimes, if several attachments included into message can not be sent as one update,
        several Telegram updates will be produced.
        Sometimes, if no text and none of the supported attachments are included,
        nothing will happen.

        :param bot: Telegram bot, that is used for connection to Telegram API.
        :param chat_id: Telegram dialog ID that the message will be sent to.
        :param message: Chatsky message that will be processed into Telegram updates.
        """

        if message.text is not None:
            await bot.send_message(
                chat_id,
                message.text,
                **grab_extra_fields(
                    message,
                    [
                        "parse_mode",
                        "disable_notification",
                        "protect_content",
                        "reply_markup",
                        "message_effect_id",
                        "reply_to_message_id",
                        "disable_web_page_preview",
                    ],
                ),
            )
        if message.attachments is not None:
            for attachment in message.attachments:
                if isinstance(attachment, Location):
                    await bot.send_location(
                        chat_id,
                        attachment.latitude,
                        attachment.longitude,
                        **grab_extra_fields(
                            attachment,
                            [
                                "horizontal_accuracy",
                                "disable_notification",
                                "protect_content",
                                "reply_markup",
                                "message_effect_id",
                                "reply_to_message_id",
                            ],
                        ),
                    )
                elif isinstance(attachment, Contact):
                    await bot.send_contact(
                        chat_id,
                        attachment.phone_number,
                        attachment.first_name,
                        attachment.last_name,
                        **grab_extra_fields(
                            attachment,
                            [
                                "vcard",
                                "disable_notification",
                                "protect_content",
                                "reply_markup",
                                "message_effect_id",
                                "reply_to_message_id",
                            ],
                        ),
                    )
                elif isinstance(attachment, Poll):
                    await bot.send_poll(
                        chat_id,
                        attachment.question,
                        [option.text for option in attachment.options],
                        **grab_extra_fields(
                            attachment,
                            [
                                "is_anonymous",
                                "type",
                                "allows_multiple_answers",
                                "correct_option_id",
                                "explanation",
                                "explanation_parse_mode",
                                "open_period",
                                "is_closed",
                                "disable_notification",
                                "protect_content",
                                "reply_markup",
                                "question_parse_mode",
                                "message_effect_id",
                                "reply_to_message_id",
                            ],
                        ),
                    )
                elif isinstance(attachment, Audio):
                    attachment_bytes = await attachment.get_bytes(self)
                    if attachment_bytes is not None:
                        await bot.send_audio(
                            chat_id,
                            attachment_bytes,
                            **grab_extra_fields(
                                attachment,
                                [
                                    "caption",
                                    "parse_mode",
                                    "performer",
                                    "title",
                                    "disable_notification",
                                    "protect_content",
                                    "reply_markup",
                                    "thumbnail",
                                    "message_effect_id",
                                    "reply_to_message_id",
                                    "filename",
                                ],
                            ),
                        )
                elif isinstance(attachment, Video):
                    attachment_bytes = await attachment.get_bytes(self)
                    if attachment_bytes is not None:
                        await bot.send_video(
                            chat_id,
                            attachment_bytes,
                            **grab_extra_fields(
                                attachment,
                                [
                                    "caption",
                                    "parse_mode",
                                    "supports_streaming",
                                    "disable_notification",
                                    "protect_content",
                                    "reply_markup",
                                    "has_spoiler",
                                    "thumbnail",
                                    "message_effect_id",
                                    "show_caption_above_media",
                                    "reply_to_message_id",
                                    "filename",
                                ],
                            ),
                        )
                elif isinstance(attachment, Animation):
                    attachment_bytes = await attachment.get_bytes(self)
                    if attachment_bytes is not None:
                        await bot.send_animation(
                            chat_id,
                            attachment_bytes,
                            **grab_extra_fields(
                                attachment,
                                [
                                    "caption",
                                    "parse_mode",
                                    "disable_notification",
                                    "protect_content",
                                    "reply_markup",
                                    "has_spoiler",
                                    "thumbnail",
                                    "message_effect_id",
                                    "show_caption_above_media",
                                    "reply_to_message_id",
                                    "filename",
                                ],
                            ),
                        )
                elif isinstance(attachment, Image):
                    attachment_bytes = await attachment.get_bytes(self)
                    if attachment_bytes is not None:
                        await bot.send_photo(
                            chat_id,
                            attachment_bytes,
                            **grab_extra_fields(
                                attachment,
                                [
                                    "caption",
                                    "parse_mode",
                                    "disable_notification",
                                    "protect_content",
                                    "reply_markup",
                                    "has_spoiler",
                                    "message_effect_id",
                                    "reply_to_message_id",
                                    "filename",
                                ],
                            ),
                        )
                elif isinstance(attachment, Sticker):
                    sticker = await attachment.get_bytes(self) if attachment.id is None else attachment.id
                    if sticker is not None:
                        await bot.send_sticker(
                            chat_id,
                            sticker,
                            **grab_extra_fields(
                                attachment,
                                [
                                    "emoji",
                                    "disable_notification",
                                    "protect_content",
                                    "reply_markup",
                                    "message_effect_id",
                                    "reply_to_message_id",
                                ],
                            ),
                        )
                elif isinstance(attachment, Document):
                    attachment_bytes = await attachment.get_bytes(self)
                    if attachment_bytes is not None:
                        await bot.send_document(
                            chat_id,
                            attachment_bytes,
                            **grab_extra_fields(
                                attachment,
                                [
                                    "caption",
                                    "parse_mode",
                                    "disable_notification",
                                    "protect_content",
                                    "reply_markup",
                                    "thumbnail",
                                    "message_effect_id",
                                    "reply_to_message_id",
                                    "filename",
                                ],
                            ),
                        )
                elif isinstance(attachment, VoiceMessage):
                    attachment_bytes = await attachment.get_bytes(self)
                    if attachment_bytes is not None:
                        await bot.send_voice(
                            chat_id,
                            attachment_bytes,
                            **grab_extra_fields(
                                attachment,
                                [
                                    "caption",
                                    "parse_mode",
                                    "disable_notification",
                                    "protect_content",
                                    "reply_markup",
                                    "message_effect_id",
                                    "reply_to_message_id",
                                    "filename",
                                ],
                            ),
                        )
                elif isinstance(attachment, VideoMessage):
                    attachment_bytes = await attachment.get_bytes(self)
                    if attachment_bytes is not None:
                        await bot.send_video_note(
                            chat_id,
                            attachment_bytes,
                            **grab_extra_fields(
                                attachment,
                                [
                                    "disable_notification",
                                    "protect_content",
                                    "reply_markup",
                                    "thumbnail",
                                    "message_effect_id",
                                    "reply_to_message_id",
                                    "filename",
                                ],
                            ),
                        )
                elif isinstance(attachment, MediaGroup):
                    files = list()
                    for media in attachment.group:
                        if isinstance(media, Image):
                            media_bytes = await media.get_bytes(self)
                            files += [
                                InputMediaPhoto(
                                    media_bytes,
                                    **grab_extra_fields(
                                        media,
                                        [
                                            "filename",
                                            "caption",
                                            "parse_mode",
                                            "has_spoiler",
                                            "show_caption_above_media",
                                        ],
                                    ),
                                ),
                            ]
                        elif isinstance(media, Video):
                            media_bytes = await media.get_bytes(self)
                            files += [
                                InputMediaVideo(
                                    media_bytes,
                                    **grab_extra_fields(
                                        media,
                                        [
                                            "filename",
                                            "caption",
                                            "parse_mode",
                                            "supports_streaming",
                                            "has_spoiler",
                                            "thumbnail",
                                            "show_caption_above_media",
                                        ],
                                    ),
                                ),
                            ]
                        elif isinstance(media, Animation):
                            media_bytes = await media.get_bytes(self)
                            files += [
                                InputMediaAnimation(
                                    media_bytes,
                                    **grab_extra_fields(
                                        media,
                                        [
                                            "filename",
                                            "caption",
                                            "parse_mode",
                                            "has_spoiler",
                                            "thumbnail",
                                            "show_caption_above_media",
                                        ],
                                    ),
                                ),
                            ]
                        elif isinstance(media, Audio):
                            media_bytes = await media.get_bytes(self)
                            files += [
                                InputMediaAudio(
                                    media_bytes,
                                    **grab_extra_fields(
                                        media,
                                        ["filename", "caption", "parse_mode", "performer", "title", "thumbnail"],
                                    ),
                                ),
                            ]
                        elif isinstance(media, Document):
                            media_bytes = await media.get_bytes(self)
                            files += [
                                InputMediaDocument(
                                    media_bytes,
                                    **grab_extra_fields(media, ["filename", "caption", "parse_mode", "thumbnail"]),
                                ),
                            ]
                        else:
                            raise ValueError(f"Attachment {type(media).__name__} can not be sent in a media group!")
                    await bot.send_media_group(
                        chat_id,
                        files,
                        **grab_extra_fields(
                            attachment,
                            [
                                "caption",
                                "disable_notification",
                                "protect_content",
                                "message_effect_id",
                                "reply_to_message_id",
                                "parse_mode",
                            ],
                        ),
                    )
                else:
                    raise ValueError(f"Attachment {type(attachment).__name__} is not supported!")

    async def _on_event(self, update: Update, _: Any, create_message: Callable[[Update], Message]) -> None:
        """
        Process Telegram update, run pipeline and send response to Telegram.

        :param update: Telegram update that will be processed.
        :param create_message: function that converts Telegram update to Chatsky message.
        """

        data_available = update.message is not None or update.callback_query is not None
        if update.effective_chat is not None and data_available:
            message = create_message(update)
            message.original_message = update
            resp = await self._pipeline_runner(message, update.effective_chat.id)
            if resp.last_response is not None:
                await self.cast_message_to_telegram_and_send(
                    self.application.bot, update.effective_chat.id, resp.last_response
                )

    async def on_message(self, update: Update, _: Any) -> None:
        """
        Process normal Telegram update, extracting Chatsky message from it
        using :py:meth:`~._AbstractTelegramInterface.extract_message_from_telegram`.

        :param update: Telegram update that will be processed.
        """

        await self._on_event(update, _, lambda s: self.extract_message_from_telegram(s.message))

    async def on_callback(self, update: Update, _: Any) -> None:
        """
        Process Telegram callback update, creating empty Chatsky message
        with only one callback query attachment from `callback_query.data` field.

        :param update: Telegram update that will be processed.
        """

        await self._on_event(
            update, _, lambda s: Message(attachments=[CallbackQuery(query_string=s.callback_query.data)])
        )

    async def connect(self, pipeline_runner: PipelineRunnerFunction, *args, **kwargs):
        self._pipeline_runner = pipeline_runner



"""
Yandex DB
---------
The Yandex DB module provides a version of the :py:class:`.DBContextStorage` class that designed to work with
Yandex and other databases. Yandex DataBase is a fully-managed cloud-native SQL service that makes it easy to set up,
operate, and scale high-performance and high-availability databases for your applications.

The Yandex DB module uses the Yandex Cloud SDK, which is a python library that allows you to work
with Yandex Cloud services using python. This allows Chatsky to easily integrate with the Yandex DataBase and
take advantage of the scalability and high-availability features provided by the service.
"""

import asyncio
import os
from typing import Hashable
from urllib.parse import urlsplit


from chatsky.script import Context

from .database import DBContextStorage
from .protocol import get_protocol_install_suggestion

try:
    import ydb
    import ydb.aio

    ydb_available = True
except ImportError:
    ydb_available = False


class YDBContextStorage(DBContextStorage):
    """
    Version of the :py:class:`.DBContextStorage` for YDB.

    :param path: Standard sqlalchemy URI string.
        When using sqlite backend in Windows, keep in mind that you have to use double backslashes '\\'
        instead of forward slashes '/' in the file path.
    :param table_name: The name of the table to use.
    """

    def __init__(self, path: str, table_name: str = "contexts", timeout=5):
        DBContextStorage.__init__(self, path)
        protocol, netloc, self.database, _, _ = urlsplit(path)
        self.endpoint = "{}://{}".format(protocol, netloc)
        self.table_name = table_name
        if not ydb_available:
            install_suggestion = get_protocol_install_suggestion("grpc")
            raise ImportError("`ydb` package is missing.\n" + install_suggestion)
        self.driver, self.pool = asyncio.run(_init_drive(timeout, self.endpoint, self.database, self.table_name))

    async def set_item_async(self, key: Hashable, value: Context):
        value = value if isinstance(value, Context) else Context.cast(value)

        async def callee(session):
            query = """
                PRAGMA TablePathPrefix("{}");
                DECLARE $queryId AS Utf8;
                DECLARE $queryContext AS Json;
                UPSERT INTO {}
                (
                    id,
                    context
                )
                VALUES
                (
                    $queryId,
                    $queryContext
                );
                """.format(
                self.database, self.table_name
            )
            prepared_query = await session.prepare(query)

            await session.transaction(ydb.SerializableReadWrite()).execute(
                prepared_query,
                {"$queryId": str(key), "$queryContext": value.model_dump_json()},
                commit_tx=True,
            )

        return await self.pool.retry_operation(callee)

    async def get_item_async(self, key: Hashable) -> Context:
        async def callee(session):
            query = """
                PRAGMA TablePathPrefix("{}");
                DECLARE $queryId AS Utf8;
                SELECT
                    id,
                    context
                FROM {}
                WHERE id = $queryId;
                """.format(
                self.database, self.table_name
            )
            prepared_query = await session.prepare(query)

            result_sets = await session.transaction(ydb.SerializableReadWrite()).execute(
                prepared_query,
                {
                    "$queryId": str(key),
                },
                commit_tx=True,
            )
            if result_sets[0].rows:
                return Context.cast(result_sets[0].rows[0].context)
            else:
                raise KeyError

        return await self.pool.retry_operation(callee)

    async def del_item_async(self, key: Hashable):
        async def callee(session):
            query = """
                PRAGMA TablePathPrefix("{}");
                DECLARE $queryId AS Utf8;
                DELETE
                FROM {}
                WHERE
                    id = $queryId
                ;
                """.format(
                self.database, self.table_name
            )
            prepared_query = await session.prepare(query)

            await session.transaction(ydb.SerializableReadWrite()).execute(
                prepared_query,
                {"$queryId": str(key)},
                commit_tx=True,
            )

        return await self.pool.retry_operation(callee)

    async def contains_async(self, key: Hashable) -> bool:
        async def callee(session):
            # new transaction in serializable read write mode
            # if query successfully completed you will get result sets.
            # otherwise exception will be raised
            query = """
                PRAGMA TablePathPrefix("{}");
                DECLARE $queryId AS Utf8;
                SELECT
                    id,
                    context
                FROM {}
                WHERE id = $queryId;
                """.format(
                self.database, self.table_name
            )
            prepared_query = await session.prepare(query)

            result_sets = await session.transaction(ydb.SerializableReadWrite()).execute(
                prepared_query,
                {
                    "$queryId": str(key),
                },
                commit_tx=True,
            )
            return len(result_sets[0].rows) > 0

        return await self.pool.retry_operation(callee)

    async def len_async(self) -> int:
        async def callee(session):
            query = """
                PRAGMA TablePathPrefix("{}");
                SELECT
                    COUNT(*) as cnt
                FROM {}
                """.format(
                self.database, self.table_name
            )
            prepared_query = await session.prepare(query)

            result_sets = await session.transaction(ydb.SerializableReadWrite()).execute(
                prepared_query,
                commit_tx=True,
            )
            return result_sets[0].rows[0].cnt

        return await self.pool.retry_operation(callee)

    async def clear_async(self):
        async def callee(session):
            query = """
                PRAGMA TablePathPrefix("{}");
                DECLARE $queryId AS Utf8;
                DELETE
                FROM {}
                ;
                """.format(
                self.database, self.table_name
            )
            prepared_query = await session.prepare(query)

            await session.transaction(ydb.SerializableReadWrite()).execute(
                prepared_query,
                {},
                commit_tx=True,
            )

        return await self.pool.retry_operation(callee)


async def _init_drive(timeout: int, endpoint: str, database: str, table_name: str):
    driver = ydb.aio.Driver(endpoint=endpoint, database=database)
    await driver.wait(fail_fast=True, timeout=timeout)

    pool = ydb.aio.SessionPool(driver, size=10)

    if not await _is_table_exists(pool, database, table_name):  # create table if it does not exist
        await _create_table(pool, database, table_name)
    return driver, pool


async def _is_table_exists(pool, path, table_name) -> bool:
    try:

        async def callee(session):
            await session.describe_table(os.path.join(path, table_name))

        await pool.retry_operation(callee)
        return True
    except ydb.SchemeError:
        return False


async def _create_table(pool, path, table_name):
    async def callee(session):
        await session.create_table(
            "/".join([path, table_name]),
            ydb.TableDescription()
            .with_column(ydb.Column("id", ydb.OptionalType(ydb.PrimitiveType.Utf8)))
            .with_column(ydb.Column("context", ydb.OptionalType(ydb.PrimitiveType.Json)))
            .with_primary_key("id"),
        )

    return await pool.retry_operation(callee)



"""
Common
------
This module contains several functions which are used to run demonstrations in tutorials.
"""

from os import getenv
from typing import Callable, Tuple, Optional, Union
from uuid import uuid4

from chatsky.script import Context, Message
from chatsky.pipeline import Pipeline
from chatsky.utils.testing.response_comparers import default_comparer


def is_interactive_mode() -> bool:  # pragma: no cover
    """
    Checking whether the tutorial code should be run in interactive mode.

    :return: `True` if it's being executed by Jupyter kernel and DISABLE_INTERACTIVE_MODE env variable isn't set,
        `False` otherwise.
    """

    shell = None
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
    finally:
        return shell != "ZMQInteractiveShell" and getenv("DISABLE_INTERACTIVE_MODE") is None


def check_happy_path(
    pipeline: Pipeline,
    happy_path: Tuple[Tuple[Union[str, Message], Union[str, Message]], ...],
    # This optional argument is used for additional processing of candidate responses and reference responses
    response_comparer: Callable[[Message, Message, Context], Optional[str]] = default_comparer,
    printout_enable: bool = True,
):
    """
    Running tutorial with provided pipeline for provided requests, comparing responses with correct expected responses.
    In cases when additional processing of responses is needed (e.g. in case of response being an HTML string),
    a special function (response comparer) is used.

    :param pipeline: The Pipeline instance, that will be used for checking.
    :param happy_path: A tuple of (request, response) tuples, so-called happy path,
        its requests are passed to pipeline and the pipeline responses are compared to its responses.
    :param response_comparer: A special comparer function that accepts received response, true response and context;
        it returns `None` is two responses are equal and transformed received response if they are different.
    :param printout_enable: A flag that enables requests and responses fancy printing (to STDOUT).
    """

    ctx_id = uuid4()  # get random ID for current context
    for step_id, (request_raw, reference_response_raw) in enumerate(happy_path):
        request = Message(text=request_raw) if isinstance(request_raw, str) else request_raw
        reference_response = (
            Message(text=reference_response_raw) if isinstance(reference_response_raw, str) else reference_response_raw
        )
        ctx = pipeline(request, ctx_id)
        candidate_response = ctx.last_response
        if printout_enable:
            print(f"(user) >>> {repr(request)}")
            print(f" (bot) <<< {repr(candidate_response)}")
        if candidate_response is None:
            raise Exception(
                f"\n\npipeline = {pipeline.info_dict}\n\n"
                f"ctx = {ctx}\n\n"
                f"step_id = {step_id}\n"
                f"request = {repr(request)}\n"
                "Candidate response is None."
            )
        parsed_response_with_deviation = response_comparer(candidate_response, reference_response, ctx)
        if parsed_response_with_deviation is not None:
            raise Exception(
                f"\n\npipeline = {pipeline.info_dict}\n\n"
                f"ctx = {ctx}\n\n"
                f"step_id = {step_id}\n"
                f"request = {repr(request)}\n"
                f"candidate_response = {repr(parsed_response_with_deviation)}\n"
                f"reference_response = {repr(reference_response)}\n"
                "candidate_response != reference_response"
            )


def run_interactive_mode(pipeline: Pipeline):  # pragma: no cover
    """
    Running tutorial with provided pipeline in interactive mode, just like with CLI messenger interface.
    The dialog won't be stored anywhere, it will only be outputted to STDOUT.

    :param pipeline: The Pipeline instance, that will be used for running.
    """

    ctx_id = uuid4()  # Random UID
    print("Start a dialogue with the bot")
    while True:
        request = input(">>> ")
        ctx = pipeline(request=Message(request), ctx_id=ctx_id)
        print(f"<<< {repr(ctx.last_response)}")



"""
Async Helpers
-------------
Tools to help with async.
"""

import asyncio
from typing import Callable, Any


async def wrap_sync_function_in_async(func: Callable, *args, **kwargs) -> Any:
    """
    Utility function, that wraps both functions and coroutines in coroutines.
    Invokes `func` if it is just a callable and awaits, if this is a coroutine.

    :param func: Callable to wrap.
    :param \\*args: Function args.
    :param \\**kwargs: Function kwargs.
    :return: What function returns.
    """
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    else:
        return func(*args, **kwargs)



"""
Pickle
------
The Pickle module provides a pickle-based version of the :py:class:`.DBContextStorage` class.
This class is used to store and retrieve context data in a pickle format.
It allows Chatsky to easily store and retrieve context data in a format that is efficient
for serialization and deserialization and can be easily used in python.

Pickle is a python library that allows to serialize and deserialize python objects.
It is efficient and fast, but it is not recommended to use it to transfer data across
different languages or platforms because it's not cross-language compatible.
"""

import asyncio
import pickle
from typing import Hashable

try:
    import aiofiles
    import aiofiles.os

    pickle_available = True
except ImportError:
    pickle_available = False

from .database import DBContextStorage, threadsafe_method
from chatsky.script import Context


class PickleContextStorage(DBContextStorage):
    """
    Implements :py:class:`.DBContextStorage` with `pickle` as driver.

    :param path: Target file URI. Example: 'pickle://file.pkl'.
    """

    def __init__(self, path: str):
        DBContextStorage.__init__(self, path)
        asyncio.run(self._load())

    @threadsafe_method
    async def len_async(self) -> int:
        return len(self.dict)

    @threadsafe_method
    async def set_item_async(self, key: Hashable, value: Context):
        self.dict.__setitem__(str(key), value)
        await self._save()

    @threadsafe_method
    async def get_item_async(self, key: Hashable) -> Context:
        await self._load()
        return Context.cast(self.dict.__getitem__(str(key)))

    @threadsafe_method
    async def del_item_async(self, key: Hashable):
        self.dict.__delitem__(str(key))
        await self._save()

    @threadsafe_method
    async def contains_async(self, key: Hashable) -> bool:
        await self._load()
        return self.dict.__contains__(str(key))

    @threadsafe_method
    async def clear_async(self):
        self.dict.clear()
        await self._save()

    async def _save(self):
        async with aiofiles.open(self.path, "wb+") as file:
            await file.write(pickle.dumps(self.dict))

    async def _load(self):
        if not await aiofiles.os.path.isfile(self.path) or (await aiofiles.os.stat(self.path)).st_size == 0:
            self.dict = dict()
            await self._save()
        else:
            async with aiofiles.open(self.path, "rb") as file:
                self.dict = pickle.loads(await file.read())



"""
Utility Functions
-----------------
The Utility Functions module contains several utility functions that are commonly used throughout Chatsky.
These functions provide a variety of utility functionality.
"""

from typing import Any, Optional, Tuple, Mapping


def _get_attrs_with_updates(
    obj: object,
    drop_attrs: Optional[Tuple[str, ...]] = None,
    replace_attrs: Optional[Mapping[str, str]] = None,
    add_attrs: Optional[Mapping[str, Any]] = None,
) -> dict:
    """
    Advanced customizable version of built-in `__dict__` property.
    Sometimes during Pipeline construction `Services` (or `ServiceGroups`) should be rebuilt,
    e.g. in case of some fields overriding.
    This method can be customized to return a dict,
    that can be spread (** operator) and passed to Service or ServiceGroup constructor.
    Base dict is formed via `vars` built-in function. All "private" or "dunder" fields are omitted.

    :param drop_attrs: A tuple of key names that should be removed from the resulting dict.
    :param replace_attrs: A mapping that should be replaced in the resulting dict.
    :param add_attrs: A mapping that should be added to the resulting dict.
    :return: Resulting dict.
    """
    drop_attrs = () if drop_attrs is None else drop_attrs
    replace_attrs = {} if replace_attrs is None else dict(replace_attrs)
    add_attrs = {} if add_attrs is None else dict(add_attrs)
    result = {}
    for attribute in vars(obj):
        if not attribute.startswith("__") and attribute not in drop_attrs:
            if attribute in replace_attrs:
                result[replace_attrs[attribute]] = getattr(obj, attribute)
            else:
                result[attribute] = getattr(obj, attribute)
    result.update(add_attrs)
    return result


def collect_defined_constructor_parameters_to_dict(**kwargs: Any):
    """
    Function, that creates dict from non-`None` constructor parameters of pipeline component.
    It is used in overriding component parameters,
    when service handler or service group service is instance of Service or ServiceGroup (or dict).
    It accepts same named parameters as component constructor.

    :return: Dict, containing key-value pairs of these parameters, that are not `None`.
    """
    return dict([(key, value) for key, value in kwargs.items() if value is not None])



"""
Responses
---------
Responses determine the response that will be sent to the user for each node of the dialog graph.
Responses are used to provide the user with information, ask questions,
or guide the conversation in a particular direction.

This module provides only one predefined response function that can be used to quickly
respond to the user and keep the conversation flowing.
"""

import random
from typing import List

from chatsky.pipeline import Pipeline
from chatsky.script import Context, Message


def choice(*responses):
    """
    Function wrapper that takes the list of responses as an input
    and returns handler which outputs a response randomly chosen from that list.

    :param responses: A list of responses for random sampling.
    """

    def choice_response_handler(ctx: Context, pipeline: Pipeline):
        return random.choice(responses)

    return choice_response_handler



"""
Types
-----
The Types module contains a set of basic data types that
are used to define the expected input and output of various components of the framework.
The types defined in this module include basic data types such as strings
and lists, as well as more complex types that are specific to the framework.
"""

from typing import Union, Callable, Tuple
from enum import Enum, auto
from typing_extensions import TypeAlias

from .keywords import Keywords

LabelType: TypeAlias = Union[str, Keywords]
"""Label can be a casual string or :py:class:`~chatsky.script.Keywords`."""
# todo: rename these to identifiers

NodeLabel1Type: TypeAlias = Tuple[str, float]
"""Label type for transitions can be `[node_name, transition_priority]`."""

NodeLabel2Type: TypeAlias = Tuple[str, str]
"""Label type for transitions can be `[flow_name, node_name]`."""

NodeLabel3Type: TypeAlias = Tuple[str, str, float]
"""Label type for transitions can be `[flow_name, node_name, transition_priority]`."""

NodeLabelTupledType: TypeAlias = Union[NodeLabel1Type, NodeLabel2Type, NodeLabel3Type]
"""Label type for transitions can be one of three different types."""
# todo: group all these types into a class

ConstLabel: TypeAlias = Union[NodeLabelTupledType, str]
"""Label functions should be annotated with this type only."""

Label: TypeAlias = Union[ConstLabel, Callable]
"""Label type for transitions should be of this type only."""

ConditionType: TypeAlias = Callable
"""Condition type can be only `Callable`."""


class ActorStage(Enum):
    """
    The class which holds keys for the handlers. These keys are used
    for the actions of :py:class:`.Actor`. Each stage represents
    a specific step in the conversation flow. Here is a brief description
    of each stage.
    """

    CONTEXT_INIT = auto()
    """
    This stage is used for the context initialization.
    It involves setting up the conversation context.
    """

    GET_PREVIOUS_NODE = auto()
    """
    This stage is used to retrieve the previous node.
    """

    REWRITE_PREVIOUS_NODE = auto()
    """
    This stage is used to rewrite the previous node.
    It involves updating the previous node in the conversation history
    to reflect any changes made during the current conversation turn.
    """

    RUN_PRE_TRANSITIONS_PROCESSING = auto()
    """
    This stage is used for running pre-transitions processing.
    It involves performing any necessary pre-processing tasks.
    """

    GET_TRUE_LABELS = auto()
    """
    This stage is used to retrieve the true labels.
    It involves determining the correct label to take based
    on the current conversation context.
    """

    GET_NEXT_NODE = auto()
    """
    This stage is used to retrieve the next node in the conversation flow.
    """

    REWRITE_NEXT_NODE = auto()
    """
    This stage is used to rewrite the next node.
    It involves updating the next node in the conversation flow
    to reflect any changes made during the current conversation turn.
    """

    RUN_PRE_RESPONSE_PROCESSING = auto()
    """
    This stage is used for running pre-response processing.
    It involves performing any necessary pre-processing tasks
    before generating the response to the user.
    """

    CREATE_RESPONSE = auto()
    """
    This stage is used for response creation.
    It involves generating a response to the user based on the
    current conversation context and any pre-processing performed.
    """

    FINISH_TURN = auto()
    """
    This stage is used for finishing the current conversation turn.
    It involves wrapping up any loose ends, such as saving context,
    before waiting for the user's next input.
    """



"""
Protocol
--------
The Protocol module contains the base code for the different communication protocols used in Chatsky.
It defines the :py:data:`.PROTOCOLS` constant, which lists all the supported protocols in Chatsky.

The module also includes a function :py:func:`.get_protocol_install_suggestion()` that is used to provide
suggestions for installing the necessary dependencies for a specific protocol.
This function takes the name of the desired protocol as an argument and returns
a string containing the necessary installation commands for that protocol.

Chatsky supports a variety of communication protocols,
which allows it to communicate with different types of databases.
"""

import json
import pathlib

with open(pathlib.Path(__file__).parent / "protocols.json", "r", encoding="utf-8") as protocols:
    PROTOCOLS = json.load(protocols)
_prtocol_keys = {"module", "class", "slug", "uri_example"}
assert all(set(proc.keys()) == _prtocol_keys for proc in PROTOCOLS.values()), "Protocols are incomplete"


def get_protocol_install_suggestion(protocol_name: str) -> str:
    """
    Provide suggestions for installing the necessary dependencies for a specific protocol.

    :param protocol_name: Protocol name.
    """
    protocol = PROTOCOLS.get(protocol_name, {})
    slug = protocol.get("slug")
    if slug:
        return f"Try to run `pip install chatsky[{slug}]`"
    return ""



"""
Report
--------
This method contains a function to print benchmark results to console.
"""

from pathlib import Path
from typing import Union, Set, Literal
import json


def report(
    file: Union[str, Path],
    display: Set[Literal["name", "desc", "config", "metrics"]] = set({"name", "metrics"}),
):
    """
    Print average results from a result file to stdout.

    Printed stats contain benchmark configs, object sizes, average benchmark values for successful cases and
    exception message for unsuccessful cases.

    :param file:
        File with benchmark results generated by
        :py:func:`~chatsky.utils.db_benchmark.benchmark.save_results_to_file`.
    :param display:
        A set of objects to display in results.
        Values allowed inside the set:

            - "name" -- displays the name of the benchmark case.
            - "desc" -- displays the description of the benchmark case.
            - "config" -- displays the config info of the benchmark case.
            - "metrics" -- displays average write, read, update read+update times.
    """
    with open(file, "r", encoding="utf-8") as fd:
        file_contents = json.load(fd)

    sep = "-" * 80

    report_result = "\n".join([sep, file_contents["name"], sep, file_contents["description"], sep, ""])

    for benchmark in file_contents["benchmarks"]:
        reported_values = {
            "name": benchmark["name"],
            "desc": benchmark["description"],
            "config": "\n".join(f"{k}: {v}" for k, v in benchmark["benchmark_config"].items()),
            "metrics": "".join(
                [
                    (
                        f"{metric.title() + ': ' + str(benchmark['average_results']['pretty_' + metric]):20}"
                        if benchmark["success"]
                        else benchmark["result"]
                    )
                    for metric in ("write", "read", "update", "read+update")
                ]
            ),
        }

        result = []
        for value_name, value in reported_values.items():
            if value_name in display:
                result.append(value)
        result.append("")

        report_result += f"\n{sep}\n".join(result)

    print(report_result, end="")



"""
Types
-----
The Types module contains special types that are used throughout `Chatsky Messengers`.
"""

from typing import Callable
from typing_extensions import TypeAlias


PollingInterfaceLoopFunction: TypeAlias = Callable[[], bool]
"""
A function type used in :py:class:`~.PollingMessengerInterface` to control polling loop.
Returns boolean (whether polling should be continued).
"""



"""
Redis
-----
The Redis module provides a Redis-based version of the :py:class:`.DBContextStorage` class.
This class is used to store and retrieve context data in a Redis.
It allows Chatsky to easily store and retrieve context data in a format that is highly scalable
and easy to work with.

Redis is an open-source, in-memory data structure store that is known for its
high performance and scalability. It stores data in key-value pairs and supports a variety of data
structures such as strings, hashes, lists, sets, and more.
Additionally, Redis can be used as a cache, message broker, and database, making it a versatile
and powerful choice for data storage and management.
"""

import json
from typing import Hashable

try:
    from redis.asyncio import Redis

    redis_available = True
except ImportError:
    redis_available = False

from chatsky.script import Context

from .database import DBContextStorage, threadsafe_method
from .protocol import get_protocol_install_suggestion


class RedisContextStorage(DBContextStorage):
    """
    Implements :py:class:`.DBContextStorage` with `redis` as the database backend.

    :param path: Database URI string. Example: `redis://user:password@host:port`.
    """

    def __init__(self, path: str):
        DBContextStorage.__init__(self, path)
        if not redis_available:
            install_suggestion = get_protocol_install_suggestion("redis")
            raise ImportError("`redis` package is missing.\n" + install_suggestion)
        self._redis = Redis.from_url(self.full_path)

    @threadsafe_method
    async def contains_async(self, key: Hashable) -> bool:
        return bool(await self._redis.exists(str(key)))

    @threadsafe_method
    async def set_item_async(self, key: Hashable, value: Context):
        value = value if isinstance(value, Context) else Context.cast(value)
        await self._redis.set(str(key), value.model_dump_json())

    @threadsafe_method
    async def get_item_async(self, key: Hashable) -> Context:
        result = await self._redis.get(str(key))
        if result:
            result_dict = json.loads(result.decode("utf-8"))
            return Context.cast(result_dict)
        raise KeyError(f"No entry for key {key}.")

    @threadsafe_method
    async def del_item_async(self, key: Hashable):
        await self._redis.delete(str(key))

    @threadsafe_method
    async def len_async(self) -> int:
        return await self._redis.dbsize()

    @threadsafe_method
    async def clear_async(self):
        await self._redis.flushdb()



from chatsky.script.core.message import (
    Location,
    Attachment,
    Audio,
    Video,
    Image,
    Document,
    # Attachments,
    # Link,
    # Button,
    # Keyboard,
    Message,
    # MultiMessage
)



"""
JSON
----
The JSON module provides a json-based version of the :py:class:`.DBContextStorage` class.
This class is used to store and retrieve context data in a JSON. It allows Chatsky to easily
store and retrieve context data.
"""

import asyncio
from typing import Hashable

try:
    import aiofiles
    import aiofiles.os

    json_available = True
except ImportError:
    json_available = False

from pydantic import BaseModel, model_validator

from .database import DBContextStorage, threadsafe_method
from chatsky.script import Context


class SerializableStorage(BaseModel, extra="allow"):
    @model_validator(mode="before")
    @classmethod
    def validate_any(cls, vals):
        for key, value in vals.items():
            vals[key] = Context.cast(value)
        return vals


class JSONContextStorage(DBContextStorage):
    """
    Implements :py:class:`.DBContextStorage` with `json` as the storage format.

    :param path: Target file URI. Example: `json://file.json`.
    """

    def __init__(self, path: str):
        DBContextStorage.__init__(self, path)
        asyncio.run(self._load())

    @threadsafe_method
    async def len_async(self) -> int:
        return len(self.storage.model_extra)

    @threadsafe_method
    async def set_item_async(self, key: Hashable, value: Context):
        self.storage.model_extra.__setitem__(str(key), value)
        await self._save()

    @threadsafe_method
    async def get_item_async(self, key: Hashable) -> Context:
        await self._load()
        return Context.cast(self.storage.model_extra.__getitem__(str(key)))

    @threadsafe_method
    async def del_item_async(self, key: Hashable):
        self.storage.model_extra.__delitem__(str(key))
        await self._save()

    @threadsafe_method
    async def contains_async(self, key: Hashable) -> bool:
        await self._load()
        return self.storage.model_extra.__contains__(str(key))

    @threadsafe_method
    async def clear_async(self):
        self.storage.model_extra.clear()
        await self._save()

    async def _save(self):
        async with aiofiles.open(self.path, "w+", encoding="utf-8") as file_stream:
            await file_stream.write(self.storage.model_dump_json())

    async def _load(self):
        if not await aiofiles.os.path.isfile(self.path) or (await aiofiles.os.stat(self.path)).st_size == 0:
            self.storage = SerializableStorage()
            await self._save()
        else:
            async with aiofiles.open(self.path, "r", encoding="utf-8") as file_stream:
                self.storage = SerializableStorage.model_validate_json(await file_stream.read())



"""
Context storage benchmarking
----------------------------
This module contains functions for context storages benchmarking.

The basic function is :py:func:`~.time_context_read_write` but it has a low level interface.

Higher level wrappers of the function provided by this module are:

- :py:func:`~.save_results_to_file` -- saves results for a list of benchmark cases.
- :py:func:`~.benchmark_all` -- a wrapper of `save_results_to_file`. Creates cases from configs.

Wrappers use :py:class:`~.BenchmarkConfig` interface to configure benchmarks.
A simple configuration class as well as a configuration set are provided by
:py:mod:`chatsky.utils.db_benchmark.basic_config`.

To view files generated by :py:func:`~.save_results_to_file` use either
:py:func:`~chatsky.utils.db_benchmark.report.report` or
`our streamlit app <../_misc/benchmark_streamlit.py>`_.
"""

from uuid import uuid4
from pathlib import Path
from time import perf_counter
from typing import Tuple, List, Dict, Union, Optional, Callable, Any
import json
import importlib
from statistics import mean
import abc
from traceback import extract_tb, StackSummary

from pydantic import BaseModel, Field
from tqdm.auto import tqdm

from chatsky.context_storages import DBContextStorage
from chatsky.script import Context


def time_context_read_write(
    context_storage: DBContextStorage,
    context_factory: Callable[[], Context],
    context_num: int,
    context_updater: Optional[Callable[[Context], Optional[Context]]] = None,
) -> Tuple[List[float], List[Dict[int, float]], List[Dict[int, float]]]:
    """
    Benchmark `context_storage` by writing and reading `context`\\s generated by `context_factory`
    into it / from it `context_num` times.
    If `context_updater` is not `None` it is used to update `context`\\s and benchmark update operation.

    This function clears `context_storage` before and after execution.

    :param context_storage: Context storage to benchmark.
    :param context_factory: A function that creates contexts which will be written into context storage.
    :param context_num: A number of times the context will be written and read.
    :param context_updater:
        None or a function.
        If not None, function should accept :py:class:`~.Context` and return an updated :py:class:`~.Context`.
        The updated context can be either the same object (at the same pointer) or a different object (e.g. copied).
        The updated context should have a higher dialog length than the received context
        (to emulate context updating during dialog).
        The function should return `None` to stop updating contexts.
        For an example of such function, see implementation of
        :py:meth:`chatsky.utils.db_benchmark.basic_config.BasicBenchmarkConfig.context_updater`.

        To avoid keeping many contexts in memory,
        this function will be called repeatedly at least `context_num` times.
    :return:
        A tuple of 3 elements.

        The first element -- a list of write times. Its length is equal to `context_num`.

        The second element -- a list of dictionaries with read times.
        Each dictionary maps from int to float. The key in the mapping is the `dialog_len` of the context and the
        values are the read times for the corresponding `dialog_len`.
        If `context_updater` is None, all dictionaries will have only one key --
        dialog length of the context returned by `context_factory`.
        Otherwise, the dictionaries will also have a key for each updated context.

        The third element -- a list of dictionaries with update times.
        Structurally the same as the second element, but none of the elements here have a key for
        dialog_len of the context returned by `context_factory`.
        So if `context_updater` is None, all dictionaries will be empty.
    """
    context_storage.clear()

    write_times: List[float] = []
    read_times: List[Dict[int, float]] = []
    update_times: List[Dict[int, float]] = []

    for _ in tqdm(range(context_num), desc=f"Benchmarking context storage:{context_storage.full_path}", leave=False):
        context = context_factory()

        ctx_id = uuid4()

        # write operation benchmark
        write_start = perf_counter()
        context_storage[ctx_id] = context
        write_times.append(perf_counter() - write_start)

        read_times.append({})
        update_times.append({})

        # read operation benchmark
        read_start = perf_counter()
        _ = context_storage[ctx_id]
        read_time = perf_counter() - read_start
        read_times[-1][len(context.labels)] = read_time

        if context_updater is not None:
            updated_context = context_updater(context)

            while updated_context is not None:
                update_start = perf_counter()
                context_storage[ctx_id] = updated_context
                update_time = perf_counter() - update_start
                update_times[-1][len(updated_context.labels)] = update_time

                read_start = perf_counter()
                _ = context_storage[ctx_id]
                read_time = perf_counter() - read_start
                read_times[-1][len(updated_context.labels)] = read_time

                updated_context = context_updater(updated_context)

        context_storage.clear()
    return write_times, read_times, update_times


class DBFactory(BaseModel):
    """
    A class for storing information about context storage to benchmark.
    Also used to create a context storage from the configuration.
    """

    uri: str
    """URI of the context storage."""
    factory_module: str = "chatsky.context_storages"
    """A module containing `factory`."""
    factory: str = "context_storage_factory"
    """Name of the context storage factory. (function that creates context storages from URIs)"""

    def db(self):
        """
        Create a context storage using `factory` from `uri`.
        """
        module = importlib.import_module(self.factory_module)
        return getattr(module, self.factory)(self.uri)


class BenchmarkConfig(BaseModel, abc.ABC, frozen=True):
    """
    Configuration for a benchmark.

    Defines methods and parameters required to run :py:func:`~.time_context_read_write`.
    Also defines a method (`info`) for displaying information about this configuration.

    A simple way to configure benchmarks is provided by
    :py:class:`~.chatsky.utils.db_benchmark.basic_config.BasicBenchmarkConfig`.

    Inherit from this class only if `BasicBenchmarkConfig` is not enough for your benchmarking needs.
    """

    context_num: int = 30
    """
    Number of times the contexts will be benchmarked.
    Increasing this number decreases standard error of the mean for benchmarked data.
    """

    @abc.abstractmethod
    def get_context(self) -> Context:
        """
        Return context to benchmark read and write operations with.

        This function will be called `context_num` times.
        """
        ...

    @abc.abstractmethod
    def info(self) -> Dict[str, Any]:
        """
        Return a dictionary with information about this configuration.
        """
        ...

    @abc.abstractmethod
    def context_updater(self, context: Context) -> Optional[Context]:
        """
        Update context with new dialog turns or return `None` to stop updates.

        This function is used to benchmark update and read operations.

        This function will be called AT LEAST `context_num` times.

        :return: Updated context or `None` to stop updating context.
        """
        ...


class BenchmarkCase(BaseModel):
    """
    This class represents a benchmark case and includes
    information about it, its configuration and configuration of a context storage to benchmark.
    """

    name: str
    """Name of a benchmark case."""
    db_factory: DBFactory
    """DBFactory that specifies context storage to benchmark."""
    benchmark_config: BenchmarkConfig
    """Benchmark configuration."""
    uuid: str = Field(default_factory=lambda: str(uuid4()))
    """Unique id of the case. Defaults to a random uuid."""
    description: str = ""
    """Description of the case. Defaults to an empty string."""

    @staticmethod
    def set_average_results(benchmark):
        """
        Modify `benchmark` dictionary to include averaged benchmark results.

        Add field "average_results" to the benchmark that contains the following fields:

            - average_write_time
            - average_read_time
            - average_update_time
            - read_times_grouped_by_context_num -- a list of read times.
              Each element is the average of read times with the same context_num.
            - read_times_grouped_by_dialog_len -- a dictionary of read times.
              Its values are the averages of read times with the same dialog_len,
              its keys are dialog_len values.
            - update_times_grouped_by_context_num
            - update_times_grouped_by_dialog_len
            - pretty_write -- average write time with only 3 significant digits.
            - pretty_read
            - pretty_update
            - pretty_read+update -- sum of average read and update times with only 3 significant digits.

        :param benchmark:
            A dictionary returned by `BenchmarkCase._run`.
            Should include a "success" and "result" fields.
            "success" field should be true.
            "result" field should be a dictionary with the values returned by
            :py:func:`~.time_context_read_write` and keys
            "write_times", "read_times" and "update_times".
        :return: None
        """
        if not benchmark["success"] or isinstance(benchmark["result"], str):
            return

        def get_complex_stats(results):
            if len(results) == 0 or len(results[0]) == 0:
                return [], {}, None

            average_grouped_by_context_num = [mean(times.values()) for times in results]
            average_grouped_by_dialog_len = {key: mean([times[key] for times in results]) for key in results[0].keys()}
            average = float(mean(average_grouped_by_context_num))
            return average_grouped_by_context_num, average_grouped_by_dialog_len, average

        read_stats = get_complex_stats(benchmark["result"]["read_times"])
        update_stats = get_complex_stats(benchmark["result"]["update_times"])

        result = {
            "average_write_time": mean(benchmark["result"]["write_times"]),
            "average_read_time": read_stats[2],
            "average_update_time": update_stats[2],
            "read_times_grouped_by_context_num": read_stats[0],
            "read_times_grouped_by_dialog_len": read_stats[1],
            "update_times_grouped_by_context_num": update_stats[0],
            "update_times_grouped_by_dialog_len": update_stats[1],
        }
        result["pretty_write"] = (
            float(f'{result["average_write_time"]:.3}') if result["average_write_time"] is not None else None
        )
        result["pretty_read"] = (
            float(f'{result["average_read_time"]:.3}') if result["average_read_time"] is not None else None
        )
        result["pretty_update"] = (
            float(f'{result["average_update_time"]:.3}') if result["average_update_time"] is not None else None
        )
        result["pretty_read+update"] = (
            float(f'{result["average_read_time"] + result["average_update_time"]:.3}')
            if result["average_read_time"] is not None and result["average_update_time"] is not None
            else None
        )

        benchmark["average_results"] = result

    def _run(self):
        try:
            write_times, read_times, update_times = time_context_read_write(
                self.db_factory.db(),
                self.benchmark_config.get_context,
                self.benchmark_config.context_num,
                self.benchmark_config.context_updater,
            )
            return {
                "success": True,
                "result": {
                    "write_times": write_times,
                    "read_times": read_times,
                    "update_times": update_times,
                },
            }
        except Exception as e:
            return {
                "success": False,
                "result": {
                    "type": e.__class__.__name__,
                    "msg": getattr(e, "message", str(e)),
                    "traceback": "\n".join(StackSummary.from_list(extract_tb(e.__traceback__)).format()),
                },
            }

    def run(self):
        """
        Run benchmark, return results.

        :return:
            A dictionary with 3 keys: "success", "result", "average_results".

            Success is a bool value. It is false if an exception was raised during benchmarking.

            Result is either an exception message or a dictionary with 3 keys
            ("write_times", "read_times", "update_times").
            Values of those fields are the values returned by :py:func:`~.time_context_read_write`.

            Average results field is as described in :py:meth:`~.BenchmarkCase.set_average_results`.
        """
        benchmark = self._run()
        BenchmarkCase.set_average_results(benchmark)
        return benchmark


def save_results_to_file(
    benchmark_cases: List[BenchmarkCase],
    file: Union[str, Path],
    name: str,
    description: str,
    exist_ok: bool = False,
):
    """
    Benchmark all `benchmark_cases` and save results to a file.

    Result are saved in json format with this schema:
    `utils/db_benchmark/benchmark_schema.json <../_misc/benchmark_schema.json>`_.

    Files created by this function cen be viewed either by using :py:func:`~chatsky.utils.db_benchmark.report.report` or
    streamlit app located in the utils directory:
    `utils/db_benchmark/benchmark_streamlit.py <../_misc/benchmark_streamlit.py>`_.

    :param benchmark_cases: A list of benchmark cases that specify benchmarks.
    :param file: File to save results to.
    :param name: Name of the benchmark set.
    :param description: Description of the benchmark set.
    :param exist_ok: Whether to continue if the file already exists.
    """
    with open(file, "w" if exist_ok else "x", encoding="utf-8") as fd:
        uuid = str(uuid4())
        result: Dict[str, Any] = {
            "name": name,
            "description": description,
            "uuid": uuid,
            "benchmarks": [],
        }
        cases = tqdm(benchmark_cases, leave=False)
        case: BenchmarkCase
        for case in cases:
            cases.set_description(f"Benchmarking: {case.name}")
            result["benchmarks"].append(
                {
                    **case.model_dump(exclude={"benchmark_config"}),
                    "benchmark_config": case.benchmark_config.info(),
                    **case.run(),
                }
            )

        json.dump(result, fd)


def benchmark_all(
    file: Union[str, Path],
    name: str,
    description: str,
    db_uri: str,
    benchmark_configs: Dict[str, BenchmarkConfig],
    exist_ok: bool = False,
):
    """
    A wrapper for :py:func:`~.save_results_to_file`.

    Generates `benchmark_cases` from `db_uri` and `benchmark_configs`:
    `db_uri` is used to initialize :py:class:`~.DBFactory` instance
    which is then used along with `benchmark_configs` to initialize :py:class:`~.BenchmarkCase` instances.

    :param file: File to save results to.
    :param name: Name of the benchmark set.
    :param description: Description of the benchmark set. The same description is used for benchmark cases.
    :param db_uri: URI of the database to benchmark
    :param benchmark_configs: Mapping from case names to configs.
    :param exist_ok: Whether to continue if the file already exists.
    """
    save_results_to_file(
        [
            BenchmarkCase(
                name=case_name,
                description=description,
                db_factory=DBFactory(uri=db_uri),
                benchmark_config=benchmark_config,
            )
            for case_name, benchmark_config in benchmark_configs.items()
        ],
        file,
        name,
        description,
        exist_ok=exist_ok,
    )



"""
Shelve
------
The Shelve module provides a shelve-based version of the :py:class:`.DBContextStorage` class.
This class is used to store and retrieve context data in a shelve format.
It allows Chatsky to easily store and retrieve context data in a format that is efficient
for serialization and deserialization and can be easily used in python.

Shelve is a python library that allows to store and retrieve python objects.
It is efficient and fast, but it is not recommended to use it to transfer data across different languages
or platforms because it's not cross-language compatible.
It stores data in a dbm-style format in the file system, which is not as fast as the other serialization
libraries like pickle or JSON.
"""

import pickle
from shelve import DbfilenameShelf
from typing import Hashable

from chatsky.script import Context

from .database import DBContextStorage


class ShelveContextStorage(DBContextStorage):
    """
    Implements :py:class:`.DBContextStorage` with `shelve` as the driver.

    :param path: Target file URI. Example: `shelve://file.db`.
    """

    def __init__(self, path: str):
        DBContextStorage.__init__(self, path)
        self.shelve_db = DbfilenameShelf(filename=self.path, protocol=pickle.HIGHEST_PROTOCOL)

    async def get_item_async(self, key: Hashable) -> Context:
        return self.shelve_db[str(key)]

    async def set_item_async(self, key: Hashable, value: Context):
        self.shelve_db.__setitem__(str(key), value)

    async def del_item_async(self, key: Hashable):
        self.shelve_db.__delitem__(str(key))

    async def contains_async(self, key: Hashable) -> bool:
        return self.shelve_db.__contains__(str(key))

    async def len_async(self) -> int:
        return self.shelve_db.__len__()

    async def clear_async(self):
        self.shelve_db.clear()



"""
Actor
-----
Actor is a component of :py:class:`.Pipeline`, that contains the :py:class:`.Script` and handles it.
It is responsible for processing user input and determining the appropriate response based
on the current state of the conversation and the script.
The actor receives requests in the form of a :py:class:`.Context` class, which contains
information about the user's input, the current state of the conversation, and other relevant data.

The actor uses the dialog graph, represented by the :py:class:`.Script` class,
to determine the appropriate response. The script contains the structure of the conversation,
including the different `nodes` and `transitions`.
It defines the possible paths that the conversation can take, and the conditions that must be met
for a transition to occur. The actor uses this information to navigate the graph
and determine the next step in the conversation.

Overall, the actor acts as a bridge between the user's input and the dialog graph,
making sure that the conversation follows the expected flow and providing a personalized experience to the user.

Below you can see a diagram of user request processing with Actor.
Both `request` and `response` are saved to :py:class:`.Context`.

.. figure:: /_static/drawio/core/user_actor.png
"""

from __future__ import annotations
import logging
import asyncio
from typing import Union, Callable, Optional, Dict, List, TYPE_CHECKING
import copy

from chatsky.utils.turn_caching import cache_clear
from chatsky.script.core.types import ActorStage, NodeLabel2Type, NodeLabel3Type, LabelType
from chatsky.script.core.message import Message

from chatsky.script.core.context import Context
from chatsky.script.core.script import Script, Node
from chatsky.script.core.normalization import normalize_label, normalize_response
from chatsky.script.core.keywords import GLOBAL, LOCAL
from chatsky.utils.devel.async_helpers import wrap_sync_function_in_async

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from chatsky.pipeline.pipeline.pipeline import Pipeline


class Actor:
    """
    The class which is used to process :py:class:`~chatsky.script.Context`
    according to the :py:class:`~chatsky.script.Script`.

    :param script: The dialog scenario: a graph described by the :py:class:`.Keywords`.
        While the graph is being initialized, it is validated and then used for the dialog.
    :param start_label: The start node of :py:class:`~chatsky.script.Script`. The execution begins with it.
    :param fallback_label: The label of :py:class:`~chatsky.script.Script`.
        Dialog comes into that label if all other transitions failed,
        or there was an error while executing the scenario.
        Defaults to `None`.
    :param label_priority: Default priority value for all :py:const:`labels <chatsky.script.ConstLabel>`
        where there is no priority. Defaults to `1.0`.
    :param condition_handler: Handler that processes a call of condition functions. Defaults to `None`.
    :param handlers: This variable is responsible for the usage of external handlers on
        the certain stages of work of :py:class:`~chatsky.script.Actor`.

        - key (:py:class:`~chatsky.script.ActorStage`) - Stage in which the handler is called.
        - value (List[Callable]) - The list of called handlers for each stage.  Defaults to an empty `dict`.
    """

    def __init__(
        self,
        script: Union[Script, dict],
        start_label: NodeLabel2Type,
        fallback_label: Optional[NodeLabel2Type] = None,
        label_priority: float = 1.0,
        condition_handler: Optional[Callable] = None,
        handlers: Optional[Dict[ActorStage, List[Callable]]] = None,
    ):
        self.script = script if isinstance(script, Script) else Script(script=script)
        self.label_priority = label_priority

        self.start_label = normalize_label(start_label)
        if self.script.get(self.start_label[0], {}).get(self.start_label[1]) is None:
            raise ValueError(f"Unknown start_label={self.start_label}")

        if fallback_label is None:
            self.fallback_label = self.start_label
        else:
            self.fallback_label = normalize_label(fallback_label)
            if self.script.get(self.fallback_label[0], {}).get(self.fallback_label[1]) is None:
                raise ValueError(f"Unknown fallback_label={self.fallback_label}")
        self.condition_handler = default_condition_handler if condition_handler is None else condition_handler

        self.handlers = {} if handlers is None else handlers

        # NB! The following API is highly experimental and may be removed at ANY time WITHOUT FURTHER NOTICE!!
        self._clean_turn_cache = True

    async def __call__(self, pipeline: Pipeline, ctx: Context):
        await self._run_handlers(ctx, pipeline, ActorStage.CONTEXT_INIT)

        # get previous node
        self._get_previous_node(ctx)
        await self._run_handlers(ctx, pipeline, ActorStage.GET_PREVIOUS_NODE)

        # rewrite previous node
        self._rewrite_previous_node(ctx)
        await self._run_handlers(ctx, pipeline, ActorStage.REWRITE_PREVIOUS_NODE)

        # run pre transitions processing
        await self._run_pre_transitions_processing(ctx, pipeline)
        await self._run_handlers(ctx, pipeline, ActorStage.RUN_PRE_TRANSITIONS_PROCESSING)

        # get true labels for scopes (GLOBAL, LOCAL, NODE)
        await self._get_true_labels(ctx, pipeline)
        await self._run_handlers(ctx, pipeline, ActorStage.GET_TRUE_LABELS)

        # get next node
        self._get_next_node(ctx)
        await self._run_handlers(ctx, pipeline, ActorStage.GET_NEXT_NODE)

        ctx.add_label(ctx.framework_data.actor_data["next_label"][:2])

        # rewrite next node
        self._rewrite_next_node(ctx)
        await self._run_handlers(ctx, pipeline, ActorStage.REWRITE_NEXT_NODE)

        # run pre response processing
        await self._run_pre_response_processing(ctx, pipeline)
        await self._run_handlers(ctx, pipeline, ActorStage.RUN_PRE_RESPONSE_PROCESSING)

        # create response
        ctx.framework_data.actor_data["response"] = await self.run_response(
            ctx.framework_data.actor_data["pre_response_processed_node"].response, ctx, pipeline
        )
        await self._run_handlers(ctx, pipeline, ActorStage.CREATE_RESPONSE)
        ctx.add_response(ctx.framework_data.actor_data["response"])

        await self._run_handlers(ctx, pipeline, ActorStage.FINISH_TURN)
        if self._clean_turn_cache:
            cache_clear()

        ctx.framework_data.actor_data.clear()

    def _get_previous_node(self, ctx: Context):
        ctx.framework_data.actor_data["previous_label"] = (
            normalize_label(ctx.last_label) if ctx.last_label else self.start_label
        )
        ctx.framework_data.actor_data["previous_node"] = self.script.get(
            ctx.framework_data.actor_data["previous_label"][0], {}
        ).get(ctx.framework_data.actor_data["previous_label"][1], Node())

    async def _get_true_labels(self, ctx: Context, pipeline: Pipeline):
        # GLOBAL
        ctx.framework_data.actor_data["global_transitions"] = (
            self.script.get(GLOBAL, {}).get(GLOBAL, Node()).transitions
        )
        ctx.framework_data.actor_data["global_true_label"] = await self._get_true_label(
            ctx.framework_data.actor_data["global_transitions"], ctx, pipeline, GLOBAL, "global"
        )

        # LOCAL
        ctx.framework_data.actor_data["local_transitions"] = (
            self.script.get(ctx.framework_data.actor_data["previous_label"][0], {}).get(LOCAL, Node()).transitions
        )
        ctx.framework_data.actor_data["local_true_label"] = await self._get_true_label(
            ctx.framework_data.actor_data["local_transitions"],
            ctx,
            pipeline,
            ctx.framework_data.actor_data["previous_label"][0],
            "local",
        )

        # NODE
        ctx.framework_data.actor_data["node_transitions"] = ctx.framework_data.actor_data[
            "pre_transitions_processed_node"
        ].transitions
        ctx.framework_data.actor_data["node_true_label"] = await self._get_true_label(
            ctx.framework_data.actor_data["node_transitions"],
            ctx,
            pipeline,
            ctx.framework_data.actor_data["previous_label"][0],
            "node",
        )

    def _get_next_node(self, ctx: Context):
        # choose next label
        ctx.framework_data.actor_data["next_label"] = self._choose_label(
            ctx.framework_data.actor_data["node_true_label"], ctx.framework_data.actor_data["local_true_label"]
        )
        ctx.framework_data.actor_data["next_label"] = self._choose_label(
            ctx.framework_data.actor_data["next_label"], ctx.framework_data.actor_data["global_true_label"]
        )
        # get next node
        ctx.framework_data.actor_data["next_node"] = self.script.get(
            ctx.framework_data.actor_data["next_label"][0], {}
        ).get(ctx.framework_data.actor_data["next_label"][1])

    def _rewrite_previous_node(self, ctx: Context):
        node = ctx.framework_data.actor_data["previous_node"]
        flow_label = ctx.framework_data.actor_data["previous_label"][0]
        ctx.framework_data.actor_data["previous_node"] = self._overwrite_node(
            node,
            flow_label,
            only_current_node_transitions=True,
        )

    def _rewrite_next_node(self, ctx: Context):
        node = ctx.framework_data.actor_data["next_node"]
        flow_label = ctx.framework_data.actor_data["next_label"][0]
        ctx.framework_data.actor_data["next_node"] = self._overwrite_node(node, flow_label)

    def _overwrite_node(
        self,
        current_node: Node,
        flow_label: LabelType,
        only_current_node_transitions: bool = False,
    ) -> Node:
        overwritten_node = copy.deepcopy(self.script.get(GLOBAL, {}).get(GLOBAL, Node()))
        local_node = self.script.get(flow_label, {}).get(LOCAL, Node())
        for node in [local_node, current_node]:
            overwritten_node.pre_transitions_processing.update(node.pre_transitions_processing)
            overwritten_node.pre_response_processing.update(node.pre_response_processing)
            overwritten_node.response = overwritten_node.response if node.response is None else node.response
            overwritten_node.misc.update(node.misc)
            if not only_current_node_transitions:
                overwritten_node.transitions.update(node.transitions)
        if only_current_node_transitions:
            overwritten_node.transitions = current_node.transitions
        return overwritten_node

    async def run_response(
        self,
        response: Optional[Union[Message, Callable[..., Message]]],
        ctx: Context,
        pipeline: Pipeline,
    ) -> Message:
        """
        Executes the normalized response as an asynchronous function.
        See the details in the :py:func:`~normalize_response` function of `normalization.py`.
        """
        response = normalize_response(response)
        return await wrap_sync_function_in_async(response, ctx, pipeline)

    async def _run_processing_parallel(self, processing: dict, ctx: Context, pipeline: Pipeline) -> None:
        """
        Execute the processing functions for a particular node simultaneously,
        independent of the order.

        Picked depending on the value of the :py:class:`.Pipeline`'s `parallelize_processing` flag.
        """
        results = await asyncio.gather(
            *[wrap_sync_function_in_async(func, ctx, pipeline) for func in processing.values()],
            return_exceptions=True,
        )
        for exc, (processing_name, processing_func) in zip(results, processing.items()):
            if isinstance(exc, Exception):
                logger.error(
                    f"Exception {exc} for processing_name={processing_name} and processing_func={processing_func}",
                    exc_info=exc,
                )

    async def _run_processing_sequential(self, processing: dict, ctx: Context, pipeline: Pipeline) -> None:
        """
        Execute the processing functions for a particular node in-order.

        Picked depending on the value of the :py:class:`.Pipeline`'s `parallelize_processing` flag.
        """
        for processing_name, processing_func in processing.items():
            try:
                await wrap_sync_function_in_async(processing_func, ctx, pipeline)
            except Exception as exc:
                logger.error(
                    f"Exception {exc} for processing_name={processing_name} and processing_func={processing_func}",
                    exc_info=exc,
                )

    async def _run_pre_transitions_processing(self, ctx: Context, pipeline: Pipeline) -> None:
        """
        Run `PRE_TRANSITIONS_PROCESSING` functions for a particular node.
        Pre-transition processing functions can modify the context state
        before the direction of the next transition is determined depending on that state.

        The execution order depends on the value of the :py:class:`.Pipeline`'s
        `parallelize_processing` flag.
        """
        ctx.framework_data.actor_data["processed_node"] = copy.deepcopy(ctx.framework_data.actor_data["previous_node"])
        pre_transitions_processing = ctx.framework_data.actor_data["previous_node"].pre_transitions_processing

        if pipeline.parallelize_processing:
            await self._run_processing_parallel(pre_transitions_processing, ctx, pipeline)
        else:
            await self._run_processing_sequential(pre_transitions_processing, ctx, pipeline)

        ctx.framework_data.actor_data["pre_transitions_processed_node"] = ctx.framework_data.actor_data[
            "processed_node"
        ]
        del ctx.framework_data.actor_data["processed_node"]

    async def _run_pre_response_processing(self, ctx: Context, pipeline: Pipeline) -> None:
        """
        Run `PRE_RESPONSE_PROCESSING` functions for a particular node.
        Pre-response processing functions can modify the response before it is
        returned to the user.

        The execution order depends on the value of the :py:class:`.Pipeline`'s
        `parallelize_processing` flag.
        """
        ctx.framework_data.actor_data["processed_node"] = copy.deepcopy(ctx.framework_data.actor_data["next_node"])
        pre_response_processing = ctx.framework_data.actor_data["next_node"].pre_response_processing

        if pipeline.parallelize_processing:
            await self._run_processing_parallel(pre_response_processing, ctx, pipeline)
        else:
            await self._run_processing_sequential(pre_response_processing, ctx, pipeline)

        ctx.framework_data.actor_data["pre_response_processed_node"] = ctx.framework_data.actor_data["processed_node"]
        del ctx.framework_data.actor_data["processed_node"]

    async def _get_true_label(
        self,
        transitions: dict,
        ctx: Context,
        pipeline: Pipeline,
        flow_label: LabelType,
        transition_info: str = "",
    ) -> Optional[NodeLabel3Type]:
        true_labels = []

        cond_booleans = await asyncio.gather(
            *(self.condition_handler(condition, ctx, pipeline) for condition in transitions.values())
        )
        for label, cond_is_true in zip(transitions.keys(), cond_booleans):
            if cond_is_true:
                if callable(label):
                    label = await wrap_sync_function_in_async(label, ctx, pipeline)
                    # TODO: explicit handling of errors
                    if label is None:
                        continue
                true_labels += [label]
        true_labels = [
            ((label[0] if label[0] else flow_label),)
            + label[1:2]
            + ((self.label_priority if label[2] == float("-inf") else label[2]),)
            for label in true_labels
        ]
        true_labels.sort(key=lambda label: -label[2])
        true_label = true_labels[0] if true_labels else None
        logger.debug(f"{transition_info} transitions sorted by priority = {true_labels}")
        return true_label

    async def _run_handlers(self, ctx, pipeline: Pipeline, actor_stage: ActorStage):
        stage_handlers = self.handlers.get(actor_stage, [])
        async_handlers = [wrap_sync_function_in_async(handler, ctx, pipeline) for handler in stage_handlers]
        await asyncio.gather(*async_handlers)

    def _choose_label(
        self, specific_label: Optional[NodeLabel3Type], general_label: Optional[NodeLabel3Type]
    ) -> NodeLabel3Type:
        if all([specific_label, general_label]):
            chosen_label = specific_label if specific_label[2] >= general_label[2] else general_label
        elif any([specific_label, general_label]):
            chosen_label = specific_label if specific_label else general_label
        else:
            chosen_label = self.fallback_label
        return chosen_label


async def default_condition_handler(
    condition: Callable, ctx: Context, pipeline: Pipeline
) -> Callable[[Context, Pipeline], bool]:
    """
    The simplest and quickest condition handler for trivial condition handling returns the callable condition:

    :param condition: Condition to copy.
    :param ctx: Context of current condition.
    :param pipeline: Pipeline we use in this condition.
    """
    return await wrap_sync_function_in_async(condition, ctx, pipeline)
